#
# Author: Tiberiu Boros
#
# Copyright (c) 2018 Adobe Systems Incorporated. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
from cube.misc.misc import fopen

sys.path.insert(0, '../')
from random import shuffle
import time
import random
from cube.misc.conll18_ud_eval_wrapper import conll_eval


# import nltk


class MTTrainer:
    def __init__(self, translator, src_enc, dst_enc, src_we, dst_we, patience, trainset, devset, testset=None):
        self.translator = translator
        self.src_enc = src_enc
        self.dst_enc = dst_enc
        self.src_we = src_we
        self.dst_we = dst_we
        self.patience = patience
        self.trainset = trainset
        self.devset = devset
        self.testset = testset

    def start_training(self, output_base, batch_size=100):
        itt_no_improve = self.patience
        selected_test_BLEU = 0
        selected_dev_BLEU = 0
        path = output_base + ".src.encodings"
        sys.stdout.write("Storing source encodings in " + path + "\n")
        self.src_enc.save(path)
        path = output_base + ".conf"
        sys.stdout.write("Storing config in " + path + "\n")
        self.translator.config.save(path)
        path = output_base + ".dst.encodings"
        sys.stdout.write("Storing destination encodings in " + path + "\n")
        self.dst_enc.save(path)
        epoch = 0
        # sys.stdout.write("\tevaluating on devset...")
        # sys.stdout.flush()
        # bleu_dev = self.eval(self.devset)
        # sys.stdout.write(" BLEU=" + str(bleu_dev) + "\n")

        while itt_no_improve > 0:
            itt_no_improve -= 1
            epoch += 1
            sys.stdout.write("Starting epoch " + str(epoch) + "\n")
            sys.stdout.write("\tshuffling training data... ")
            sys.stdout.flush()
            shuffle(self.trainset.sequences)
            sys.stdout.write("done\n")
            sys.stdout.flush()
            sys.stdout.write("\ttraining...")
            sys.stdout.flush()
            total_loss = 0
            start_time = time.time()
            cbs = 0
            self.translator.start_batch()
            last_proc = 0
            for iSeq in range(len(self.trainset.sequences)):
                seq = self.trainset.sequences[iSeq]
                proc = int((iSeq + 1) * 100 / len(self.trainset.sequences))
                if proc % 5 == 0 and proc != last_proc:
                    last_proc = proc
                    sys.stdout.write(" " + str(proc))
                    sys.stdout.flush()

                self.translator.learn(seq.src, seq.dst)
                cbs += len(seq.src) + len(seq.dst)
                if cbs >= batch_size:
                    total_loss += self.translator.end_batch()
                    self.translator.start_batch()

            total_loss += self.translator.end_batch()
            stop_time = time.time()
            sys.stdout.write(" avg_loss=" + str(total_loss / len(self.trainset.sequences)) + " execution_time=" + str(
                stop_time - start_time) + "\n")

            sys.stdout.write("\tevaluating on devset...")
            sys.stdout.flush()
            bleu_dev = self.eval(self.devset, filename=output_base + ".dev.out")
            sys.stdout.write(" BLEU=" + str(bleu_dev) + "\n")

            if self.testset is not None:
                sys.stdout.write("\tevaluating on testset...")
                sys.stdout.flush()
                bleu_test = self.eval(self.testset, filename=output_base + ".test.out")
                sys.stdout.write(" BLEU=" + str(bleu_test) + "\n")
            else:
                bleu_test = 0
            if bleu_dev > selected_dev_BLEU:
                selected_dev_BLEU = bleu_dev
                selected_test_BLEU = bleu_test
                itt_no_improve = self.patience
                self.translator.save(output_base + ".bestBLEU")
            self.translator.save(output_base + ".last")

        sys.stdout.write("Training is done with devset BLEU=" + str(selected_dev_BLEU) + " and testset BLEU=" + str(
            selected_test_BLEU) + " (for the selected devset)\n")

    def eval(self, dataset, filename=None):
        total_bleu = 0.0
        last_proc = 0
        iSeq = 0
        if filename is not None:
            f = fopen(filename, "w", encoding="utf-8")

        for seq in dataset.sequences:
            proc = int((iSeq + 1) * 100 / len(dataset.sequences))
            if proc % 5 == 0 and proc != last_proc:
                last_proc = proc
                sys.stdout.write(" " + str(proc))
                sys.stdout.flush()
            iSeq += 1

            hyp = self.translator.translate(seq.src)
            ref = [entry.word for entry in seq.dst]
            hyp = list(hyp)
            ref = list(ref)
            # print "hyp=",hyp
            # print "ref=",ref
            # print "\n\n\n\n"
            # sys.stdout.flush()
            if filename is not None:
                for entry in seq.src:
                    f.write(entry.word + " ")
                f.write("\n")
                for entry in seq.dst:
                    f.write(entry.word + " ")
                f.write("\n")
                for word in hyp:
                    f.write(word.encode('utf-8') + " ")
                f.write("\n\n")

            if len(ref) >= 4 and len(hyp) >= 4:
                score = nltk.translate.bleu_score.sentence_bleu([ref], hyp)
                total_bleu += score
        if filename is not None:
            f.close()
        return total_bleu / len(dataset.sequences)


class LemmatizerTrainer:
    def __init__(self, lemmatizer, encodings, patience, trainset, devset, testset=None):
        self.tagger = lemmatizer
        self.trainset = trainset
        self.devset = devset
        self.testset = testset
        self.patience = patience
        self.encodings = encodings

    def start_training(self, output_base, batch_size=1):
        epoch = 0
        itt_no_improve = self.patience
        selected_test_acc = 0
        selected_dev_acc = 0
        path = output_base + ".encodings"
        sys.stdout.write("Storing encodings in " + path + "\n")
        self.encodings.save(path)
        path = output_base + ".conf"
        sys.stdout.write("Storing config in " + path + "\n")
        self.tagger.config.save(path)
        sys.stdout.write("\tevaluating on devset...")
        sys.stdout.flush()
        dev_acc = 0  # self.eval(self.devset)
        sys.stdout.write(" accuracy=" + str(dev_acc) + "\n")
        if self.testset is not None:
            sys.stdout.write("\tevaluating on testset...")
            sys.stdout.flush()
            test_acc = 0  # self.eval(self.testset)
            sys.stdout.write(" accuracy=" + str(test_acc) + "\n")
        best_dev_acc = dev_acc

        while itt_no_improve > 0:
            itt_no_improve -= 1
            epoch += 1
            sys.stdout.write("Starting epoch " + str(epoch) + "\n")
            sys.stdout.flush()
            sys.stdout.write("\tshuffling training data... ")
            sys.stdout.flush()
            shuffle(self.trainset.sequences)
            sys.stdout.write("done\n")
            sys.stdout.flush()
            last_proc = 0
            sys.stdout.write("\ttraining...")
            sys.stdout.flush()
            total_loss = 0
            start_time = time.time()
            current_batch_size = 0
            self.tagger.start_batch()
            for iSeq in range(len(self.trainset.sequences)):
                seq = self.trainset.sequences[iSeq]
                proc = int((iSeq + 1) * 100 / len(self.trainset.sequences))
                if proc % 5 == 0 and proc != last_proc:
                    last_proc = proc
                    sys.stdout.write(" " + str(proc))
                    sys.stdout.flush()

                self.tagger.learn(seq)
                current_batch_size += len(seq)
                if current_batch_size >= batch_size:
                    total_loss += self.tagger.end_batch()
                    self.tagger.start_batch()
                    current_batch_size = 0
            total_loss += self.tagger.end_batch()
            stop_time = time.time()
            sys.stdout.write(" avg_loss=" + str(total_loss / len(self.trainset.sequences)) + " execution_time=" + str(
                stop_time - start_time) + "\n")

            sys.stdout.write("\tevaluating on devset...")
            sys.stdout.flush()
            dev_acc = self.eval(self.devset)
            sys.stdout.write(" accuracy=" + str(dev_acc) + "\n")
            if self.testset is not None:
                sys.stdout.write("\tevaluating on testset...")
                sys.stdout.flush()
                test_acc = self.eval(self.testset)
                sys.stdout.write(" accuracy=" + str(test_acc) + ")\n")

            if dev_acc > best_dev_acc:
                best_dev_acc = dev_acc
                selected_dev_acc = dev_acc
                if self.testset is not None:
                    selected_test_acc = test_acc
                path = output_base + ".bestAcc"
                sys.stdout.write("\tStoring " + path + "\n")
                sys.stdout.flush()
                self.tagger.save(path)
                itt_no_improve = self.patience

            path = output_base + ".last"
            sys.stdout.write("\tStoring " + path + "\n")
            sys.stdout.flush()
            self.tagger.save(path)

        sys.stdout.write("Training is done with devset_accuracy=" + str(selected_dev_acc) + "\n")
        if self.testset is not None:
            sys.stdout.write(" and testset_accuracy=" + str(
                selected_test_acc) + " (for the selected epoch, based on best devset accuracy)")
        sys.stdout.write("\n")

    def eval(self, dataset):
        last_proc = 0
        correct = 0
        total = 0

        for iSeq in range(len(dataset.sequences)):
            seq = dataset.sequences[iSeq]

            proc = int((iSeq + 1) * 100 / len(dataset.sequences))
            if proc % 5 == 0 and proc != last_proc:
                last_proc = proc
                sys.stdout.write(" " + str(proc))
                sys.stdout.flush()

            pred_lemmas = self.tagger.tag(seq)

            for entry, pred_lemma in zip(seq, pred_lemmas):
                if entry.upos != 'NUM' and entry.upos != 'PROPN':
                    total += 1
                    # from pdb import set_trace
                    # set_trace()
                    if sys.version_info[0] == 2:
                        if unicode(entry.lemma, 'utf-8') == pred_lemma:
                            correct += 1
                    else:
                        if entry.lemma == pred_lemma:
                            correct += 1
                else:
                    correct += 1
                    total += 1

        return float(correct) / total


class CompoundWordTrainer:
    def __init__(self, cw, encodings, patience, trainset, devset, testset=None):
        self.tagger = cw
        self.trainset = trainset
        self.devset = devset
        self.testset = testset
        self.patience = patience
        self.encodings = encodings

    def start_training(self, output_base, batch_size=1):
        epoch = 0
        itt_no_improve = self.patience
        selected_test_acc = 0
        selected_dev_acc = 0
        path = output_base + ".encodings"
        sys.stdout.write("Storing encodings in " + path + "\n")
        self.encodings.save(path)
        path = output_base + ".conf"
        sys.stdout.write("Storing config in " + path + "\n")
        self.tagger.config.save(path)
        sys.stdout.write("\tevaluating on devset...")
        sys.stdout.flush()
        dev_fscore = 0  # self.eval(self.devset)
        dev_acc = 0
        sys.stdout.write(" accuracy=" + str(dev_acc) + "\n")
        if self.testset is not None:
            sys.stdout.write("\tevaluating on testset...")
            sys.stdout.flush()
            test_fscore = 0
            test_acc = 0  # self.eval(self.testset)
            sys.stdout.write(" accuracy=" + str(test_acc) + "\n")
        best_dev_acc = dev_acc

        while itt_no_improve > 0:
            itt_no_improve -= 1
            epoch += 1
            sys.stdout.write("Starting epoch " + str(epoch) + "\n")
            sys.stdout.flush()
            sys.stdout.write("\tshuffling training data... ")
            sys.stdout.flush()
            shuffle(self.trainset.sequences)
            sys.stdout.write("done\n")
            sys.stdout.flush()
            last_proc = 0
            sys.stdout.write("\ttraining...")
            sys.stdout.flush()
            total_loss = 0
            start_time = time.time()
            current_batch_size = 0
            self.tagger.start_batch()
            for iSeq in range(len(self.trainset.sequences)):
                seq = self.trainset.sequences[iSeq]
                proc = int((iSeq + 1) * 100 / len(self.trainset.sequences))
                if proc % 5 == 0 and proc != last_proc:
                    last_proc = proc
                    sys.stdout.write(" " + str(proc))
                    sys.stdout.flush()

                self.tagger.learn(seq)
                current_batch_size += len(seq)
                if current_batch_size >= batch_size:
                    total_loss += self.tagger.end_batch()
                    self.tagger.start_batch()
                    current_batch_size = 0
            total_loss += self.tagger.end_batch()
            stop_time = time.time()
            sys.stdout.write(" avg_loss=" + str(total_loss / len(self.trainset.sequences)) + " execution_time=" + str(
                stop_time - start_time) + "\n")

            sys.stdout.write("\tevaluating on devset...")
            sys.stdout.flush()
            dev_fscore, dev_acc = self.eval(self.devset)
            sys.stdout.write(" fscore=" + str(dev_fscore) + " accuracy=" + str(dev_acc) + "\n")
            if self.testset is not None:
                sys.stdout.write("\tevaluating on testset...")
                sys.stdout.flush()
                test_fscore, test_acc = self.eval(self.testset)
                sys.stdout.write(" fscore=" + str(test_fscore) + " accuracy=" + str(test_acc) + ")\n")

            if dev_acc > best_dev_acc:
                best_dev_acc = dev_acc
                selected_dev_acc = dev_acc
                selected_dev_fscore = dev_fscore
                if self.testset is not None:
                    selected_test_acc = test_acc
                    selected_test_fscore = test_fscore
                path = output_base + ".bestAcc"
                sys.stdout.write("\tStoring " + path + "\n")
                sys.stdout.flush()
                self.tagger.save(path)
                itt_no_improve = self.patience

            path = output_base + ".last"
            sys.stdout.write("\tStoring " + path + "\n")
            sys.stdout.flush()
            self.tagger.save(path)

        sys.stdout.write(
            "Training is done with devset fscore=" + str(selected_dev_fscore) + " acc=" + str(selected_dev_acc) + "\n")
        if self.testset is not None:
            sys.stdout.write(" and testset fscore=" + str(
                selected_test_fscore) + " acc=" + str(
                selected_test_acc) + " (for the selected epoch, based on best devset accuracy)")
        sys.stdout.write("\n")

    def eval(self, dataset):
        detection_correct = 0
        detection_incorrect = 0
        detection_total = 0
        detection_real = 0

        tokens_correct = 0
        tokens_total = 0

        last_proc = 0

        for iSeq in range(len(dataset.sequences)):
            seq = dataset.sequences[iSeq]

            proc = int((iSeq + 1) * 100 / len(dataset.sequences))
            if proc % 5 == 0 and proc != last_proc:
                last_proc = proc
                sys.stdout.write(" " + str(proc))
                sys.stdout.flush()

            i_entry = 0
            while i_entry < len(seq):
                entry = seq[i_entry]
                if entry.is_compound_entry:
                    detection_real += 1

                    compound, tokens = self.tagger.tag_token(entry.word)
                    if compound:
                        detection_correct += 1
                        detection_total += 1
                    interval = entry.index.split("-")
                    interval = int(interval[1]) - int(interval[0]) + 1
                    real_tokens = []
                    for _ in range(interval):
                        i_entry += 1
                        real_tokens.append(seq[i_entry].word)
                    i_entry += 1
                    tokens_total += len(real_tokens)
                    for pt, rt in zip(tokens, real_tokens):
                        if sys.version_info[0] == 2:
                            if pt.encode('utf-8') == rt:
                                tokens_correct += 1
                        else:
                            if pt == rt:
                                tokens_correct += 1

                else:
                    compound, _ = self.tagger.tag_token(entry.word)
                    if compound:
                        detection_incorrect += 1
                        detection_total += 1
                i_entry += 1
        if detection_total == 0:
            detection_total += 1
        if detection_real == 0:
            detection_real += 1
        p = float(detection_correct) / detection_total
        r = float(detection_correct) / detection_real
        if p == 0 or r == 0:
            f = 0
        else:
            f = 2 * p * r / (p + r)

        acc = float(tokens_correct) / tokens_total
        return f, acc


class TaggerTrainer:
    def __init__(self, tagger, encodings, patience, trainset, devset, testset=None):
        self.tagger = tagger
        self.trainset = trainset
        self.devset = devset
        self.testset = testset
        self.patience = patience
        self.encodings = encodings

    def start_training(self, output_base, batch_size=0):
        epoch = 0
        itt_no_improve = self.patience
        selected_test_upos, selected_test_xpos, selected_test_attrs = 0, 0, 0
        selected_dev_upos, selected_dev_xpos, selected_dev_attrs = 0, 0, 0
        path = output_base + ".encodings"
        sys.stdout.write("Storing encodings in " + path + "\n")
        self.encodings.save(path)
        path = output_base + ".conf"
        sys.stdout.write("Storing config in " + path + "\n")
        self.tagger.config.save(path)
        sys.stdout.write("\tevaluating on devset...")
        sys.stdout.flush()
        dev_upos, dev_xpos, dev_attrs = self.eval(self.devset)
        sys.stdout.write(" accuracy=( UPOS=" + str(dev_upos) + " , XPOS=" + str(dev_xpos) + " , ATTRS=" + str(
            dev_attrs) + " )\n")
        if self.testset is not None:
            sys.stdout.write("\tevaluating on testset...")
            sys.stdout.flush()
            test_acc = self.eval(self.testset)
            sys.stdout.write(" accuracy=" + str(test_acc) + "\n")
        best_dev_upos = dev_upos
        best_dev_xpos = dev_xpos
        best_dev_attrs = dev_attrs
        best_dev_overall = (dev_upos + dev_xpos + dev_attrs) / 3

        while itt_no_improve > 0:
            itt_no_improve -= 1
            epoch += 1
            sys.stdout.write("Starting epoch " + str(epoch) + "\n")
            sys.stdout.flush()
            sys.stdout.write("\tshuffling training data... ")
            sys.stdout.flush()
            shuffle(self.trainset.sequences)
            sys.stdout.write("done\n")
            sys.stdout.flush()
            last_proc = 0
            sys.stdout.write("\ttraining...")
            sys.stdout.flush()
            total_loss = 0
            start_time = time.time()
            current_batch_size = 0
            self.tagger.start_batch()
            for iSeq in range(len(self.trainset.sequences)):
                seq = self.trainset.sequences[iSeq][0]
                lang_id = self.trainset.sequences[iSeq][1]
                proc = int((iSeq + 1) * 100 / len(self.trainset.sequences))
                if proc % 5 == 0 and proc != last_proc:
                    last_proc = proc
                    sys.stdout.write(" " + str(proc))
                    sys.stdout.flush()

                self.tagger.learn(seq, lang_id=lang_id)
                current_batch_size += len(seq)
                if current_batch_size > batch_size:
                    total_loss += self.tagger.end_batch()
                    self.tagger.start_batch()
                    current_batch_size = 0

            if current_batch_size != 0:
                total_loss += self.tagger.end_batch()
                self.tagger.start_batch()

            stop_time = time.time()
            sys.stdout.write(" avg_loss=" + str(total_loss / len(self.trainset.sequences)) + " execution_time=" + str(
                stop_time - start_time) + "\n")

            sys.stdout.write("\tevaluating on trainset...")
            sys.stdout.flush()
            train_upos, train_xpos, train_attrs = self.eval(self.trainset)
            sys.stdout.write(" accuracy=( UPOS=" + str(train_upos) + " , XPOS=" + str(train_xpos) + " , ATTRS=" + str(
                train_attrs) + " )\n")

            sys.stdout.write("\tevaluating on devset...")
            sys.stdout.flush()
            dev_upos, dev_xpos, dev_attrs = self.eval(self.devset)
            sys.stdout.write(" accuracy=( UPOS=" + str(dev_upos) + " , XPOS=" + str(dev_xpos) + " , ATTRS=" + str(
                dev_attrs) + " )\n")
            if self.testset is not None:
                sys.stdout.write("\tevaluating on testset...")
                sys.stdout.flush()
                test_upos, test_xpos, test_attrs = self.eval(self.testset)
                sys.stdout.write(" accuracy=( UPOS=" + str(test_upos) + " , XPOS=" + str(test_xpos) + " , ATTRS=" + str(
                    test_attrs) + " )\n")

            if (dev_upos + dev_xpos + dev_attrs) / 3 > best_dev_overall:
                best_dev_overall = (dev_upos + dev_xpos + dev_attrs) / 3
                selected_dev_upos = dev_upos
                selected_dev_xpos = dev_xpos
                selected_dev_attrs = dev_attrs
                if self.testset is not None:
                    selected_test_upos = test_upos
                    selected_test_xpos = test_xpos
                    selected_test_attrs = test_attrs
                path = output_base + ".bestOVERALL"
                sys.stdout.write("\tStoring " + path + "\n")
                sys.stdout.flush()
                self.tagger.save(path)
                itt_no_improve = self.patience

            if dev_upos > best_dev_upos:
                best_dev_upos = dev_upos
                path = output_base + ".bestUPOS"
                sys.stdout.write("\tStoring " + path + "\n")
                sys.stdout.flush()
                self.tagger.save(path)
                itt_no_improve = self.patience

            if dev_xpos > best_dev_xpos:
                best_dev_xpos = dev_xpos
                path = output_base + ".bestXPOS"
                sys.stdout.write("\tStoring " + path + "\n")
                sys.stdout.flush()
                self.tagger.save(path)
                itt_no_improve = self.patience

            if dev_attrs > best_dev_attrs:
                best_dev_attrs = dev_attrs
                path = output_base + ".bestATTRS"
                sys.stdout.write("\tStoring " + path + "\n")
                sys.stdout.flush()
                self.tagger.save(path)
                itt_no_improve = self.patience

            path = output_base + ".last"
            sys.stdout.write("\tStoring " + path + "\n")
            sys.stdout.flush()
            self.tagger.save(path)

        sys.stdout.write("Training is done with devset accuracy=( UPOS=" + str(selected_dev_upos) + " , XPOS=" + str(
            selected_dev_xpos) + " , ATTRS=" + str(selected_dev_attrs) + " )\n")
        if self.testset is not None:
            sys.stdout.write(" and testset_accuracy=( UPOS=" + str(selected_test_upos) + " , XPOS=" + str(
                selected_test_xpos) + " , ATTRS=" + str(
                selected_test_attrs) + " ) (for the selected epoch, based on best devset accuracy)")
        sys.stdout.write("\n")

    def eval(self, dataset):
        last_proc = 0
        correct_upos = 0
        correct_xpos = 0
        correct_attrs = 0
        total = 0

        for iSeq in range(len(dataset.sequences)):
            seq = dataset.sequences[iSeq][0]
            lang_id = dataset.sequences[iSeq][1]

            proc = int((iSeq + 1) * 100 / len(dataset.sequences))
            if proc % 5 == 0 and proc != last_proc:
                last_proc = proc
                sys.stdout.write(" " + str(proc))
                sys.stdout.flush()

            pred_tags = self.tagger.tag(seq, lang_id=lang_id)

            for entry, pred_tag in zip(seq, pred_tags):
                total += 1

                if entry.upos == pred_tag[0]:
                    correct_upos += 1
                if entry.xpos == pred_tag[1]:
                    correct_xpos += 1
                if entry.attrs == pred_tag[2]:
                    correct_attrs += 1

        return float(correct_upos) / total, float(correct_xpos) / total, float(correct_attrs) / total


class ParserTrainer:
    def __init__(self, parser, encodings, patience, trainset, devset, testset=None):
        self.parser = parser
        self.trainset = trainset
        self.devset = devset
        self.testset = testset
        self.patience = patience
        self.encodings = encodings

    def start_training(self, output_base, batch_size=100):

        epoch = 0
        itt_no_improve = self.patience

        path = output_base + ".encodings"
        sys.stdout.write("Storing encodings in " + path + "\n")
        self.encodings.save(path)
        path = output_base + ".conf"
        sys.stdout.write("Storing config in " + path + "\n")
        self.parser.config.save(path)
        # sys.stdout.write("\tevaluating on devset...")
        # sys.stdout.flush()
        # dev_uas, dev_las, dev_upos, dev_xpos, dev_attrs, dev_lemma = self.eval(self.devset)
        # sys.stdout.write(" UAS=" + str(dev_uas) + " LAS=" + str(dev_las) + " UPOS=" + str(dev_upos) + " XPOS=" + str(
        #     dev_xpos) + " ATTRS=" + str(dev_attrs) + " LEMMA=" + str(dev_lemma) + "\n")
        # if self.testset is not None:
        #     sys.stdout.write("\tevaluating on testset...")
        #     sys.stdout.flush()
        #     test_uas, test_las, test_upos, test_xpos, test_attrs, test_lemma = self.eval(self.testset)
        #     sys.stdout.write(
        #         " UAS=" + str(test_uas) + " LAS=" + str(test_las) + " UPOS=" + str(test_upos) + " XPOS=" + str(
        #             test_xpos) + " ATTRS=" + str(test_attrs) + " LEMMA=" + str(test_lemma) + "\n")

        # best_dev_uas = dev_uas
        # best_dev_las = dev_las
        best_dev_uas = 0
        best_dev_las = 0
        test_uas_uas = 0
        test_uas_las = 0
        test_las_uas = 0
        test_las_las = 0
        dev_uas_uas = 0
        dev_uas_las = 0
        dev_las_uas = 0
        dev_las_las = 0
        current_batch_size = 0
        self.parser.start_batch()

        while itt_no_improve > 0:

            itt_no_improve -= 1
            epoch += 1
            sys.stdout.write("Starting epoch " + str(epoch) + "\n")
            sys.stdout.flush()
            sys.stdout.write("\tshuffling training data... ")
            sys.stdout.flush()
            shuffle(self.trainset.sequences)
            sys.stdout.write("done\n")
            sys.stdout.flush()
            last_proc = 0
            sys.stdout.write("\ttraining...")
            sys.stdout.flush()
            total_loss = 0
            start_time = time.time()

            for iSeq in range(len(self.trainset.sequences)):
                seq = self.trainset.sequences[iSeq][0]
                lang_id = self.trainset.sequences[iSeq][1]
                proc = int((iSeq + 1) * 100 / len(self.trainset.sequences))
                if proc % 5 == 0 and proc != last_proc:
                    last_proc = proc
                    sys.stdout.write(" " + str(proc))
                    sys.stdout.flush()

                self.parser.learn(seq, lang_id)
                current_batch_size += len(seq)
                if current_batch_size >= batch_size:
                    total_loss += self.parser.end_batch()
                    current_batch_size = 0
                    self.parser.start_batch()
            total_loss += self.parser.end_batch()
            current_batch_size = 0
            stop_time = time.time()
            sys.stdout.write(" avg_loss=" + str(total_loss / len(self.trainset.sequences)) + " execution_time=" + str(
                stop_time - start_time) + "\n")
            self.parser.start_batch()

            # sys.stdout.write("\tevaluating on trainset...")
            # sys.stdout.flush()
            # train_uas, train_las = self.eval(self.trainset)
            # sys.stdout.write(" UAS=" + str(train_uas) + " LAS=" + str(train_las) + "\n")

            sys.stdout.write("\tevaluating on devset...")
            sys.stdout.flush()
            dev_uas, dev_las, dev_upos, dev_xpos, dev_attrs = self.eval(self.devset)
            sys.stdout.write(
                " UAS=" + str(dev_uas) + " LAS=" + str(dev_las) + " UPOS=" + str(dev_upos) + " XPOS=" + str(
                    dev_xpos) + " ATTRS=" + str(dev_attrs) + "\n")
            if self.testset is not None:
                sys.stdout.write("\tevaluating on testset...")
                sys.stdout.flush()
                test_uas, test_las, test_upos, test_xpos, test_attrs = self.eval(self.testset)
                sys.stdout.write(
                    " UAS=" + str(test_uas) + " LAS=" + str(test_las) + " UPOS=" + str(test_upos) + " XPOS=" + str(
                        test_xpos) + " ATTRS=" + str(test_attrs) + "\n")

            if dev_uas > best_dev_uas:
                best_dev_uas = dev_uas
                dev_uas_uas = dev_uas
                dev_uas_las = dev_las
                if self.testset is not None:
                    test_uas_uas = test_uas
                    test_uas_las = test_las
                path = output_base + ".bestUAS"
                sys.stdout.write("\tStoring " + path + "\n")
                sys.stdout.flush()
                self.parser.save(path)
                itt_no_improve = self.patience

            if dev_las > best_dev_las:
                best_dev_las = dev_las
                dev_las_uas = dev_uas
                dev_las_las = dev_las
                if self.testset is not None:
                    test_las_uas = test_uas
                    test_las_las = test_las
                path = output_base + ".bestLAS"
                sys.stdout.write("\tStoring " + path + "\n")
                sys.stdout.flush()
                self.parser.save(path)
                itt_no_improve = self.patience

            path = output_base + ".last"
            sys.stdout.write("\tStoring " + path + "\n")
            sys.stdout.flush()
            self.parser.save(path)

        sys.stdout.write("Training is done with devset\n")
        sys.stdout.write("Best UAS score provides:\n")
        sys.stdout.write("\tDev UAS=" + str(dev_uas_uas) + " LAS=" + str(dev_uas_las) + "\n")
        sys.stdout.write("\tTest UAS=" + str(test_uas_uas) + " LAS=" + str(test_uas_las) + "\n")
        sys.stdout.write("Best LAS score provides:\n")
        sys.stdout.write("\tDev UAS=" + str(dev_las_uas) + " LAS=" + str(dev_las_las) + "\n")
        sys.stdout.write("\tTest UAS=" + str(test_las_uas) + " LAS=" + str(test_las_las) + "\n")
        sys.stdout.write("\n")

    def eval(self, dataset):
        last_proc = 0
        correct_uas = 0
        correct_las = 0
        correct_upos = 0
        correct_xpos = 0
        correct_attrs = 0

        total = 0
        for iSeq in range(len(dataset.sequences)):
            seq = dataset.sequences[iSeq][0]
            lang_id = dataset.sequences[iSeq][1]
            # remove compound words
            tmp = []
            for entry in seq:
                if not entry.is_compound_entry:
                    tmp.append(entry)
            seq = tmp
            proc = int((iSeq + 1) * 100 / len(dataset.sequences))
            if proc % 5 == 0 and proc != last_proc:
                last_proc = proc
                sys.stdout.write(" " + str(proc))
                sys.stdout.flush()

            predicted = self.parser.tag(seq, lang_id)

            for entry, pred in zip(seq, predicted):
                total += 1
                gold_head = entry.head
                gold_label = entry.label
                pred_head = pred.head
                pred_label = pred.label

                if pred_head == gold_head:
                    correct_uas += 1
                    if gold_label == pred_label:
                        correct_las += 1

                if pred.upos == entry.upos:
                    correct_upos += 1
                if pred.xpos == entry.xpos:
                    correct_xpos += 1
                if pred.attrs == entry.attrs:
                    correct_attrs += 1

        if total == 0:
            total += 1
        return float(correct_uas) / total, float(correct_las) / total, float(correct_upos) / total, float(
            correct_xpos) / total, float(correct_attrs) / total


class TokenizerTrainer:
    def __init__(self, tokenizer, encodings, patience, trainset, devset):
        self.tokenizer = tokenizer
        self.encodings = encodings
        self.patience = patience
        self.trainset = trainset
        self.devset = devset

        self.train_buckets = {}
        self.dev_buckets = {}

        self._bucket_languages()

    def _bucket_languages(self):
        max_lang_id = 0
        for seq, lang_id in self.trainset.sequences:
            if lang_id > max_lang_id:
                max_lang_id = lang_id

        for lang_id in range(max_lang_id + 1):
            self.train_buckets[lang_id] = []
            self.dev_buckets[lang_id] = []

        for seq, lang_id in self.trainset.sequences:
            self.train_buckets[lang_id].append(seq)

        for seq, lang_id in self.devset.sequences:
            self.dev_buckets[lang_id].append(seq)

    def _make_input(self, seqs):
        chars = []
        subtokens_left = 0
        for seq in seqs:
            for entry in seq:
                if subtokens_left == 0:
                    for char_idx in range(len(entry.word)):
                        chars.append(entry.word[char_idx])
                    if "spaceafter=no" not in entry.space_after.lower():
                        chars.append(' ')
                else:
                    subtokens_left -= 1
                if entry.is_compound_entry:
                    parts = entry.index.split('-')
                    subtokens_left = int(parts[1]) - int(parts[0]) + 1
        return ''.join(chars)

    def eval(self, output_base):
        # todo: implement multilanguage training

        pred_file = open(output_base + "-temporary.conllu", 'w')
        gold_file = open(output_base + "-gold.conllu", 'w')
        for lang_id in self.dev_buckets:
            text = self._make_input(self.dev_buckets[lang_id])
            seqs = self.tokenizer.tokenize(text, lang_id=lang_id)
            # with fopen(output_base + "-temporary.conllu", 'w') as file:
            for sentence in seqs:
                # print ("Sentence has entries: "+str(len(sentence)))
                for entry in sentence:
                    line = str(
                        entry.index) + "\t" + entry.word + "\t" + entry.lemma + "\t" + entry.upos + "\t" + entry.xpos + "\t" + entry.attrs + "\t" + str(
                        entry.head) + "\t" + entry.label + "\t" + entry.deps + "\t" + entry.space_after + "\n"
                    pred_file.write(line)

                pred_file.write("\n")

            for sentence in self.dev_buckets[lang_id]:
                # print ("Sentence has entries: "+str(len(sentence)))
                for entry in sentence:
                    line = str(
                        entry.index) + "\t" + entry.word + "\t" + entry.lemma + "\t" + entry.upos + "\t" + entry.xpos + "\t" + entry.attrs + "\t" + str(
                        entry.head) + "\t" + entry.label + "\t" + entry.deps + "\t" + entry.space_after + "\n"
                    gold_file.write(line)

                gold_file.write("\n")

            # run eval script

        pred_file.close()
        gold_file.close()
        metrics = conll_eval(output_base + "-temporary.conllu", output_base + "-gold.conllu")

        # return metrics["Tokens"].f1 * 100., metrics["Sentences"].f1 * 100.
        if metrics is None:
            return 0, 0, 0
        else:
            return metrics["Sentences"].f1, metrics["Tokens"].f1, metrics["Words"].f1
        # return 0, 0, 0

    def _get_num_chars(self, seq):
        seq = seq[0]
        num_chars = 0
        for entry in seq:
            num_chars += len(entry.word)
            if "spaceafter=no" not in entry.space_after.lower():
                num_chars += 1
        return num_chars

    def start_training(self, output_base, batch_size=1000):
        best_sent = 0
        best_token = 0
        best_word = 0
        patience_left = self.patience
        total_loss = 0
        total_chars = 0
        path = output_base + ".encodings"
        sys.stdout.write("Storing encodings in " + path + "\n")
        self.encodings.save(path)
        path = output_base + ".conf"
        sys.stdout.write("Storing config in " + path + "\n")
        self.tokenizer.config.save(path)

        # print(self.eval("tmp"))

        # toto: multilnaguage training
        epoch = 1
        while patience_left > 0:
            sys.stdout.write('Starting epoch ' + str(epoch) + '\n')
            epoch += 1
            patience_left -= 1
            last_proc = 0
            sys.stdout.write('\tShuffling training data\n')
            random.shuffle(self.trainset.sequences)
            sys.stdout.write('\ttraining...')
            sys.stdout.flush()

            chars_in_batch = 0
            batched_seqs = []
            start = time.time()
            total_seqs = len(self.trainset.sequences)
            curr_seqs = 0
            for lang_id in self.train_buckets:
                for idx in range(len(self.train_buckets[lang_id])):
                    curr_seqs += 1
                    curr_proc = curr_seqs * 100 // total_seqs
                    # print(curr_proc)
                    if curr_proc % 5 == 0 and last_proc != curr_proc:
                        while last_proc < curr_proc:
                            last_proc += 5
                            sys.stdout.write(' ' + str(last_proc))
                            sys.stdout.flush()

                    batched_seqs.append(self.train_buckets[lang_id][idx])
                    chars_in_batch += self._get_num_chars(self.trainset.sequences[idx])
                    if chars_in_batch > batch_size:
                        loss = self.tokenizer.learn(batched_seqs, lang_id=lang_id)
                        total_loss += loss
                        total_chars += chars_in_batch
                        chars_in_batch = 0
                        batched_seqs = []

                if chars_in_batch != 0:
                    loss = self.tokenizer.learn(batched_seqs, lang_id=lang_id)
                    total_loss += loss
                    total_chars += chars_in_batch
                    chars_in_batch = 0
                    batched_seqs = []

                stop = time.time()

            sys.stdout.write(' loss=' + str(total_loss / total_chars) + ' execution time=' + str(stop - start) + '\n')

            sys.stdout.write('\tevaluating...')
            f_sent, f_token, f_word = self.eval(output_base)
            sys.stdout.write(' sent=' + str(f_sent) + ' tok=' + str(f_token) + ' words=' + str(f_word) + '\n')
            path = output_base + ".last"
            sys.stdout.write('\t\tStoring ' + path + '\n')
            self.tokenizer.save(path)

            if f_sent > best_sent:
                best_sent = f_sent
                patience_left = self.patience
                sys.stdout.write('\t\tStoring ' + output_base + '-ss.best\n')
                self.tokenizer.save(output_base + '-ss.best')

            if f_token > best_token:
                best_token = f_token
                patience_left = self.patience
                sys.stdout.write('\t\tStoring ' + output_base + '-tok.best\n')
                self.tokenizer.save(output_base + '-tok.best')

            if f_word > best_word:
                best_word = f_word
                patience_left = self.patience
                sys.stdout.write('\t\tStoring ' + output_base + '-words.best\n')
                self.tokenizer.save(output_base + '-words.best')


class TieredTokenizerTrainer:
    def __init__(self, tokenizer, encodings, patience, trainset, devset=None, testset=None, raw_train_file=None,
                 raw_dev_file=None, raw_test_file=None, gold_train_file=None, gold_dev_file=None, gold_test_file=None):
        self.tokenizer = tokenizer
        self.trainset = trainset
        self.devset = devset
        self.raw_train_file = raw_train_file
        self.raw_dev_file = raw_dev_file
        self.raw_test_file = raw_test_file
        self.gold_train_file = gold_train_file
        self.gold_dev_file = gold_dev_file
        self.gold_test_file = gold_test_file
        self.testset = testset
        self.patience = patience
        self.encodings = encodings

    # creates from a Dataset.sequences (which is a list of lists of Conll entries)
    # a two lists of lists of chars and gold labels
    def _create_Xy_sequences(self, sequence_set):
        X_set = []
        y_set = []
        space_after_end_of_sentence = True
        space_after_end_of_sentence_count = 0
        for i in range(len(sequence_set.sequences)):  # for all sequences (sentences)
            X = []
            y = []
            # print(">> Starting sequence "+str(i))
            for j in range(len(sequence_set.sequences[i])):  # for each word in the sentence
                if sequence_set.sequences[i][j].is_compound_entry:  # skip over compound words
                    pass
                word = sequence_set.sequences[i][j].word
                space_after = False if "SpaceAfter=No" in sequence_set.sequences[i][j].space_after else True
                if sys.version_info[0] == 2:
                    uniword = unicode(word, 'utf-8')
                else:
                    import copy
                    uniword = copy.deepcopy(word)
                # print("  WORD: "+uniword+" len = "+str(len(uniword))+" space after = "+str(sequence_set.sequences[i][j].space_after))
                # mark all symbols as "O"
                for char_index in range(len(uniword)):
                    X.append(uniword[char_index])
                    y.append("O")
                    # last symbol always is "S"
                y[-1] = "S"
                # is this the last symbol in the sentence? 
                if j == len(sequence_set.sequences[i]) - 1:  # yes, last symbol in sentence, add an "X" to its label
                    y[-1] = "SX"
                    # permanently set space_after_end_of_sentence to False if we see just one sentence that does not have a space after
                    if not space_after:
                        space_after_end_of_sentence_count += 1
                else:  # add space after only if not end of sentence
                    if space_after:
                        X.append(" ")
                        y.append("S")
                        # for q in range(len(X)):
            #    print X[q]+" "+y[q]
            # raw_input("Stop")
            X_set.append(X)
            y_set.append(y)

        # if at least 75% of sentences end with space_after="No" then we assume this language does not use spaces after EOSes.
        if float(space_after_end_of_sentence_count) > 0.75 * len(sequence_set.sequences):
            space_after_end_of_sentence = False

        return X_set, y_set, space_after_end_of_sentence

    def _create_mixed_sequences(self, X_set, y_set, space_after_end_of_sentence, shuffle=False):
        assert (len(X_set) == len(y_set))
        # print(" Set has "+str(len(X_set))+" sequences")
        X_mixed_set = []
        y_mixed_set = []
        for i in range(len(X_set)):
            import copy
            X_mixed = copy.deepcopy(X_set[i])
            y_mixed = copy.deepcopy(y_set[i])
            if space_after_end_of_sentence:
                X_mixed.append(" ")
                y_mixed.append("S")
                # now, add random sentence
            pick = random.randint(0, len(X_set) - 1)
            X_mixed = X_mixed + copy.deepcopy(X_set[pick])
            y_mixed = y_mixed + copy.deepcopy(y_set[pick])
            if space_after_end_of_sentence:
                X_mixed.append(" ")
                y_mixed.append("S")
                # add some random chars of another sentence
            while True:  # some sequences have only one word, skip them
                pick = random.randint(0, len(X_set) - 1)
                if len(X_set[pick]) > 1:
                    break
            char_count = random.randint(1, len(X_set[pick]) - 1)
            X_mixed = X_mixed + copy.deepcopy(X_set[pick][0:char_count])
            y_mixed = y_mixed + copy.deepcopy(y_set[pick][0:char_count])

            # for q in range(len(X_mixed)):
            #    print X_mixed[q]+" "+y_mixed[q]
            # raw_input("Stop")
            X_mixed_set.append(X_mixed)
            y_mixed_set.append(y_mixed)

        if shuffle:
            combined = list(zip(X_mixed_set, y_mixed_set))
            random.shuffle(combined)
            X_mixed_set[:], y_mixed_set[:] = zip(*combined)

        return X_mixed_set, y_mixed_set

    def start_training(self, output_base, batch_size=0):
        epoch = 0
        itt_no_improve = self.patience
        best_dev_tok = 0.
        best_dev_ss = 0.
        selected_test_tok = 0.
        selected_test_ss = 0.
        path = output_base + ".encodings"
        sys.stdout.write("Storing encodings in " + path + "\n")
        self.encodings.save(path)
        path = output_base + ".conf"
        sys.stdout.write("Storing config in " + path + "\n")
        self.tokenizer.config.save(path)
        """
        sys.stdout.write("\tevaluating on devset...")
        sys.stdout.flush()
        dev_acc = self.eval(self.devset)
        sys.stdout.write(" accuracy=" + str(dev_acc) + "\n")
        if self.testset is not None:
            sys.stdout.write("\tevaluating on testset...")
            sys.stdout.flush()
            test_acc = self.eval(self.testset)
            sys.stdout.write(" accuracy=" + str(test_acc) + "\n")
        best_dev_acc = dev_acc
        """
        # convert Dataset to list of chars
        X_train_raw, y_train_raw, space_after_end_of_sentence = self._create_Xy_sequences(self.trainset)
        if not space_after_end_of_sentence:
            print("\t NOTE: Training sentences do not end with a space after EOS.")
        # X_dev_raw, y_dev_raw, _ = self._create_Xy_sequences(self.devset)

        while itt_no_improve > 0:
            itt_no_improve -= 1
            epoch += 1
            sys.stdout.write("Starting epoch " + str(epoch) + "\n")
            sys.stdout.flush()
            sys.stdout.write("\tshuffling training data... ")
            sys.stdout.flush()
            X_train, y_train = self._create_mixed_sequences(X_train_raw, y_train_raw, space_after_end_of_sentence,
                                                            shuffle=True)
            sys.stdout.write("done\n")
            sys.stdout.flush()

            last_proc = 0
            sys.stdout.write("\ttraining...")
            sys.stdout.flush()
            total_loss = 0
            start_time = time.time()
            current_batch_size = 0
            self.tokenizer.start_batch()
            for iSeq in range(len(X_train)):
                # print("TRAIN SEQ: "+str(iSeq))
                X = X_train[iSeq]
                y = y_train[iSeq]
                current_batch_size += len(X)
                proc = int((iSeq + 1) * 100 / len(X_train))
                if proc % 5 == 0 and proc != last_proc:
                    last_proc = proc
                    sys.stdout.write(" " + str(proc))
                    sys.stdout.flush()

                self.tokenizer.learn_ss(X, y)
                self.tokenizer.learn_tok(X, y)
                if current_batch_size >= batch_size:
                    current_batch_size = 0
                    total_loss += self.tokenizer.end_batch()
                    self.tokenizer.start_batch()
            if current_batch_size != 0:
                current_batch_size = 0
                total_loss += self.tokenizer.end_batch()
                self.tokenizer.start_batch()

            stop_time = time.time()
            sys.stdout.write(" avg_loss=" + str(total_loss / len(X_train)) + " execution_time=" + str(
                stop_time - start_time) + "\n")

            # sys.stdout.write("\tevaluating on trainset...")
            # sys.stdout.flush()

            # train_tok, train_ss = self.eval(self.raw_train_file, self.gold_train_file)
            # train_tok, train_ss = self.eval(self.raw_test_file, self.gold_test_file)
            # sys.stdout.write(" token accuracy=" + str(train_tok) + " , sentence accuracy=" + str(train_ss) + "\n")

            sys.stdout.write("\tevaluating on devset...")
            sys.stdout.flush()
            dev_tok, dev_ss = self.eval(self.raw_dev_file, self.gold_dev_file)
            sys.stdout.write(" token_accuracy=" + str(dev_tok) + " , sentence_accuracy=" + str(dev_ss) + "\n")
            if self.testset is not None:
                sys.stdout.write("\tevaluating on testset...")
                sys.stdout.flush()
                test_tok, test_ss = self.eval(self.raw_test_file, self.gold_test_file)
                sys.stdout.write(" token_accuracy=" + str(test_tok) + " , sentence_accuracy=" + str(test_ss) + "\n")
            if dev_ss > best_dev_ss:
                best_dev_ss = dev_ss
                if self.testset is not None:
                    selected_test_ss = test_ss
                path = output_base + "-ss.bestAcc"
                sys.stdout.write("\tStoring " + path + "\n")
                sys.stdout.flush()
                self.tokenizer.save_ss(path)
                itt_no_improve = self.patience
            if dev_tok > best_dev_tok:
                best_dev_tok = dev_tok
                if self.testset is not None:
                    selected_test_tok = test_tok
                path = output_base + "-tok.bestAcc"
                sys.stdout.write("\tStoring " + path + "\n")
                sys.stdout.flush()
                self.tokenizer.save_tok(path)
                itt_no_improve = self.patience

            path = output_base + "-ss.last"
            sys.stdout.write("\tStoring " + path + "\n")
            sys.stdout.flush()
            self.tokenizer.save_ss(path)
            path = output_base + "-tok.last"
            sys.stdout.write("\tStoring " + path + "\n")
            sys.stdout.flush()
            self.tokenizer.save_ss(path)

        sys.stdout.write(
            "Training is done with devset sentence tok = " + str(best_dev_tok) + " and sentence = " + str(best_dev_ss))
        if self.testset is not None:
            sys.stdout.write(
                " and testset sentence tok = " + str(selected_test_tok) + " and sentence = " + str(selected_test_ss) +
                "(for the selected epoch, based on best devset tok/ss accuracy)")
        sys.stdout.write("\n")

    def eval(self, raw_text_file, gold_conllu_file):
        input_string = ""
        useSpaces = " "  # True
        lines = []

        with fopen(raw_text_file, "r") as file:
            lines = file.readlines()

        # analyze use of spaces in first part of the file
        test = "";
        cnt = 0
        while True:
            test = test + lines[cnt]
            # print(lines[cnt])
            cnt += 1
            if cnt >= len(lines) or cnt > 5:
                break

        if float(test.count(' ')) / float(len(test)) < 0.02:
            useSpaces = ""
        # print (str(float(test.count(' '))/float(len(test))))

        i = -1
        input_string = ""
        sentences = []
        while i < len(lines) - 1:
            i += 1
            input_string = input_string + lines[i].replace("\r", "").replace("\n", "").strip() + useSpaces
            if lines[i].strip() == "" or i == len(lines) - 1:  # end of block
                if input_string.strip() != "":
                    sentences += self.tokenizer.tokenize(input_string)
                input_string = ""

        with fopen(self.tokenizer.config.base + "-temporary.conllu", 'w') as file:
            for sentence in sentences:
                # print ("Sentence has entries: "+str(len(sentence)))
                for entry in sentence:
                    line = str(
                        entry.index) + "\t" + entry.word + "\t" + entry.lemma + "\t" + entry.upos + "\t" + entry.xpos + "\t" + entry.attrs + "\t" + str(
                        entry.head) + "\t" + entry.label + "\t" + entry.deps + "\t" + entry.space_after + "\n"
                    file.write(line)

                file.write("\n")

        # run eval script
        metrics = conll_eval(self.tokenizer.config.base + "-temporary.conllu", gold_conllu_file)

        return metrics["Tokens"].f1 * 100., metrics["Sentences"].f1 * 100.


class NERTrainer:
    def __init__(self, ner, encodings, patience, trainset, devset, testset=None):
        self.ner = ner
        self.trainset = trainset
        self.devset = devset
        self.testset = testset
        self.patience = patience
        self.encodings = encodings

    def start_training(self, output_base, batch_size=100):

        epoch = 0
        itt_no_improve = self.patience

        path = output_base + ".encodings"
        sys.stdout.write("Storing encodings in " + path + "\n")
        self.encodings.save(path)
        path = output_base + ".conf"
        sys.stdout.write("Storing config in " + path + "\n")
        self.ner.config.save(path)
        sys.stdout.write("\tevaluating on devset...")
        sys.stdout.flush()

        best_dev_score = 0
        best_dev_precision = 0
        best_dev_recall = 0
        best_test_score = 0
        best_test_precision = 0
        best_test_recall = 0

        current_batch_size = 0
        self.ner.start_batch()

        while itt_no_improve > 0:

            itt_no_improve -= 1
            epoch += 1
            sys.stdout.write("Starting epoch " + str(epoch) + "\n")
            sys.stdout.flush()
            sys.stdout.write("\tshuffling training data... ")
            sys.stdout.flush()
            shuffle(self.trainset.sequences)
            sys.stdout.write("done\n")
            sys.stdout.flush()
            last_proc = 0
            sys.stdout.write("\ttraining...")
            sys.stdout.flush()
            total_loss = 0
            start_time = time.time()

            for iSeq in range(len(self.trainset.sequences)):
                seq = self.trainset.sequences[iSeq]
                proc = int((iSeq + 1) * 100 / len(self.trainset.sequences))
                if proc % 5 == 0 and proc != last_proc:
                    while last_proc < proc:
                        last_proc += 5
                        sys.stdout.write(" " + str(last_proc))
                        sys.stdout.flush()

                self.ner.learn(seq)
                current_batch_size += len(seq)
                if current_batch_size >= batch_size:
                    total_loss += self.ner.end_batch()
                    current_batch_size = 0
                    self.ner.start_batch()
            total_loss += self.ner.end_batch()
            current_batch_size = 0
            stop_time = time.time()
            sys.stdout.write(" avg_loss=" + str(total_loss / len(self.trainset.sequences)) + " execution_time=" + str(
                stop_time - start_time) + "\n")
            self.ner.start_batch()

            sys.stdout.write("\tevaluating on devset...")
            sys.stdout.flush()
            dev_precision, dev_recall, dev_score = self.eval(self.devset)
            sys.stdout.write(
                " P=" + str(dev_precision) + " R=" + str(dev_recall) + " F=" + str(dev_score) + "\n")
            if self.testset is not None:
                sys.stdout.write("\tevaluating on testset...")
                sys.stdout.flush()
                test_precision, test_recall, test_score = self.eval(self.testset)
                sys.stdout.write(
                    " P=" + str(test_precision) + " R=" + str(test_recall) + " F=" + str(test_score) + "\n")

            if dev_score > best_dev_score:
                best_dev_score = dev_score
                best_dev_precision = dev_precision
                best_dev_recall = dev_recall

                if self.testset is not None:
                    best_test_score = test_score
                    best_test_precision = test_precision
                    best_test_recall = test_recall
                path = output_base + ".bestFScore"
                sys.stdout.write("\tStoring " + path + "\n")
                sys.stdout.flush()
                self.ner.save(path)
                itt_no_improve = self.patience

            path = output_base + ".last"
            sys.stdout.write("\tStoring " + path + "\n")
            sys.stdout.flush()
            self.ner.save(path)

        sys.stdout.write("Training is done with devset\n")
        sys.stdout.write("Best UAS score provides:\n")
        sys.stdout.write(
            "\tDev P=" + str(best_dev_precision) + " R=" + str(best_dev_recall) + " F=" + str(best_dev_score) + "\n")
        if self.testset is not None:
            sys.stdout.write("\tTest P=" + str(best_test_precision) + " R=" + str(best_test_recall) + " F=" + str(
                best_test_score) + "\n")
        sys.stdout.write("\n")

    def eval(self, dataset):
        true_p = 0
        false_p = 0
        total_p = 0
        last_proc = 0
        iSeq = 0

        for s in dataset.sequences:
            iSeq += 1
            proc = int(iSeq * 100 / len(dataset.sequences))
            if proc % 15 == 0 and last_proc != proc:
                while last_proc < proc:
                    last_proc += 5
                    sys.stdout.write(" " + str(last_proc))
                    sys.stdout.flush()

            import dynet as dy
            dy.renew_cg()  # This is a special case for trainers. We evaluate the graph itself instead of the final output

            output, proj_x = self.ner._predict(s, runtime=True)
            for iSrc in range(len(s)):
                for iDst in range(len(s)):
                    if iDst > iSrc:
                        # from network import get_link
                        from generic_networks.ner import get_link
                        link = get_link(s, iSrc, iDst)
                        if link == 1:
                            total_p += 1
                        p_val = output[iSrc][iDst].value()
                        if p_val >= 0.5:
                            link_pred = 1
                        else:
                            link_pred = 0

                        if link_pred == 1 and link == 1:
                            true_p += 1
                        elif link_pred == 1:
                            false_p += 1

        # print(" ", true_p, false_p, total_p)
        if false_p + true_p == 0:
            false_p += 1
        precision = float(true_p) / float(false_p + true_p)
        if total_p == 0:
            total_p += 1
        recall = float(true_p) / total_p
        if precision == 0 or recall == 0:
            fscore = 0
        else:
            fscore = float(2 * precision * recall) / (precision + recall)

        return precision, recall, fscore



import sys
import tqdm
from abc import abstractmethod

import torch
import numpy as np
from typing import *

sys.path.append('')
from transformers import AutoTokenizer
from transformers import AutoModel
from cube.io_utils.objects import Sentence, Document
import fasttext
import fasttext.util


class LMHelper:
    def __init__(self):
        pass

    @abstractmethod
    def get_embedding_size(self):
        pass

    @abstractmethod
    def apply(self, document: Document):
        pass

    @abstractmethod
    def apply_raw(self, batch):
        pass


class LMHelperFT(LMHelper):
    def __init__(self, device: str = 'cpu', model: str = None):
        from pathlib import Path
        home = str(Path.home())
        filename = '{0}/.fasttext/cc.{1}.300.bin'.format(home, model)
        import os
        if not os.path.exists(filename):
            fasttext.util.download_model(model, if_exists='ignore')  # English
            in_file = "cc.{0}.300.bin".format(model)
            import shutil
            import pathlib
            print("Creating " + "{0}/.fasttext/".format(home))
            pathlib.Path("{0}/.fasttext/".format(home)).mkdir(parents=True, exist_ok=True)
            shutil.move(in_file, filename)
        self._fasttext = fasttext.load_model(filename)

    def get_embedding_size(self):
        return [300]

    def apply(self, document: Document):
        for ii in tqdm.tqdm(range(len(document.sentences)), desc="Pre-computing embeddings", unit="sent"):
            for jj in range(len(document.sentences[ii].words)):
                document.sentences[ii].words[jj].emb = [self._fasttext.get_word_vector(
                    document.sentences[ii].words[jj].word)]

    def apply_raw(self, batch):
        embeddings = []
        for ii in range(len(batch)):
            c_emb = []
            for jj in range(len(batch[ii])):
                c_emb.append(self._fasttext.get_word_vector(batch[ii][jj]))
            embeddings.append(c_emb)
        return embeddings


class LMHelperLanguasito(LMHelper):
    def __init__(self, device: str = 'cpu', model: str = None):
        if model is None:
            print("UserWarning: No languasito model was specified. Instance will fail")
        from languasito.api import LanguasitoAPI
        self._languasito = LanguasitoAPI.load(model)
        self._languasito.to(device)

    def get_embedding_size(self):
        # TODO: a better way to get the embedding size (right now it is hardcoded)
        return [1024]

    def apply(self, document: Document):
        BATCH_SIZE = 8
        num_batches = len(document.sentences) // BATCH_SIZE
        if len(document.sentences) % BATCH_SIZE != 0:
            num_batches += 1

        for iBatch in tqdm.tqdm(range(num_batches), desc="Pre-computing embeddings", unit="sent"):
            start = iBatch * BATCH_SIZE
            stop = min(iBatch * BATCH_SIZE + BATCH_SIZE, len(document.sentences))
            batch = []
            for ii in range(start, stop):
                cb = []
                for w in document.sentences[ii].words:
                    cb.append(w.word)
                batch.append(cb)
            embeddings = self._languasito(batch)
            for ii in range(len(batch)):
                for jj in range(len(batch[ii])):
                    document.sentences[ii + start].words[jj].emb = [embeddings[ii][jj]]

    def apply_raw(self, batch):
        BATCH_SIZE = 8
        num_batches = len(batch) // BATCH_SIZE
        if len(batch) % BATCH_SIZE != 0:
            num_batches += 1

        for iBatch in range(num_batches):
            start = iBatch * BATCH_SIZE
            stop = min(iBatch * BATCH_SIZE + BATCH_SIZE, len(batch))
            tb = []
            for ii in range(start, stop):
                cb = []
                for w in batch[ii]:
                    cb.append(w)
                tb.append(cb)
            embeddings = self._languasito(batch)

        return embeddings


class LMHelperHF(LMHelper):
    def __init__(self, device: str = 'cpu', model: str = None):
        if model is None:
            self._splitter = AutoTokenizer.from_pretrained('xlm-roberta-base')
            self._xlmr = AutoModel.from_pretrained('xlm-roberta-base',
                                                   output_hidden_states=True)
        else:
            self._splitter = AutoTokenizer.from_pretrained(model)
            self._xlmr = AutoModel.from_pretrained(model, output_hidden_states=True)
        self._xlmr.eval()
        self._xlmr.to(device)
        self._device = device
        tmp = self._xlmr(torch.tensor([[100]], device=device))
        h_state_size = tmp['hidden_states'][0].shape[-1]
        self._emb_size = [h_state_size for _ in range(len(tmp['hidden_states']))]

    def get_embedding_size(self):
        # TODO: a better way to get the embedding size (right now it is hardcoded)
        return self._emb_size

    def _compute_we(self, batch: [Sentence]):
        # XML-Roberta

        # convert all words into wordpiece indices
        word2pieces = {}
        new_sents = []
        START = 0
        PAD = 1
        END = 2
        for ii in range(len(batch)):
            c_sent = [START]
            pos = 1
            for jj in range(len(batch[ii].words)):
                word = batch[ii].words[jj].word
                pieces = self._splitter(word)['input_ids'][1:-1]
                word2pieces[(ii, jj)] = []
                for piece in pieces:
                    c_sent.append(piece)
                    word2pieces[(ii, jj)].append([ii, pos])
                    pos += 1
            c_sent.append(END)
            new_sents.append(c_sent)
        max_len = max([len(s) for s in new_sents])
        input_ids = np.ones((len(new_sents), max_len), dtype=np.long) * PAD  # pad everything
        for ii in range(input_ids.shape[0]):
            for jj in range(input_ids.shape[1]):
                if jj < len(new_sents[ii]):
                    input_ids[ii, jj] = new_sents[ii][jj]
        with torch.no_grad():
            x = torch.tensor(input_ids, device=self._device)
            max_s_len = x.shape[1]
            count = max_s_len // 512

            if max_s_len % 512 != 0:
                count += 1
            we_list = []
            for index in range(count):
                out = self._xlmr(x[:, index * 512:min(x.shape[1], index * 512 + 512)], return_dict=True)
                we = torch.cat(out['hidden_states'], dim=-1).detach().cpu()
                we_list.append(we)
            we = torch.cat(we_list, dim=1).numpy()

        word_emb = []
        for ii in range(len(batch)):
            for jj in range(len(batch[ii].words)):
                pieces = word2pieces[ii, jj]
                if len(pieces) != 0:
                    m = we[pieces[0][0], pieces[0][1]]
                    for zz in range(len(pieces) - 1):
                        m += we[pieces[zz][0], pieces[zz][1]]
                    m = m / len(pieces)
                else:
                    m = np.zeros((768 * 13), dtype=np.float)
                word_emb.append(m)
        # word_emb = torch.cat(word_emb, dim=0)

        return word_emb

    def apply(self, doc: Document):
        import tqdm
        for sent in doc.sentences:  # tqdm.tqdm(doc.sentences, desc="Pre-computing embeddings", unit="sent"):
            wemb = self._compute_we([sent])
            for ii in range(len(wemb)):
                ww = wemb[ii]
                www = []
                for kk in range(13):
                    www.append(ww[kk * 768:kk * 768 + 768])
                sent.words[ii].emb = www

    def apply_raw(self, batch):
        pass


if __name__ == "__main__":
    from ipdb import set_trace

    set_trace()



import sys
import random
from abc import abstractmethod

sys.path.append('')
import numpy as np
import torch
from torch.utils.data.dataset import Dataset

from cube.io_utils.objects import Document, Sentence, Token, Word
from cube.io_utils.encodings import Encodings

from collections import namedtuple
from transformers import AutoModel, AutoTokenizer


def unpack(data: torch.Tensor, sizes: [], max_size: int, device: str):
    pos = 0
    blist = []
    for ii in range(len(sizes)):
        slist = []
        for jj in range(sizes[ii]):
            slist.append(data[pos, :].unsqueeze(0).to(device))
            pos += 1

        for jj in range(max_size - sizes[ii]):
            slist.append(torch.zeros((1, data.shape[-1]), device=device, dtype=torch.float))
        slist = torch.cat(slist, dim=0)
        blist.append(slist.unsqueeze(0))

    blist = torch.cat(blist, dim=0)
    return blist


def mask_concat(representations, drop_prob: float, training: bool, device: str):
    if training:
        masks = []
        for ii in range(len(representations)):
            mask = np.ones((representations[ii].shape[0], representations[ii].shape[1]), dtype=np.long)
            masks.append(mask)

        for ii in range(masks[0].shape[0]):
            for jj in range(masks[0].shape[1]):
                mult = 1
                for kk in range(len(masks)):
                    p = random.random()
                    if p < drop_prob:
                        mult += 1
                        masks[kk][ii, jj] = 0
                for kk in range(len(masks)):
                    masks[kk][ii, jj] *= mult
        for kk in range(len(masks)):
            masks[kk] = torch.tensor(masks[kk], device=device)

        for kk in range(len(masks)):
            representations[kk] = representations[kk] * masks[kk].unsqueeze(2)

    return torch.cat(representations, dim=-1)


class TokenizationDataset(Dataset):
    def __init__(self, document: Document, shuffle=True):
        self._document = document
        self._shuffle = shuffle

    def __len__(self):
        return len(self._document.sentences)

    def __getitem__(self, item):
        # append two random sentences
        if self._shuffle:
            index1 = random.randint(0, len(self._document.sentences) - 1)
            index2 = random.randint(0, len(self._document.sentences) - 1)
        else:
            index1 = item - 1
            index2 = item + 1
        if index1 >= 0:
            prev = self._document.sentences[index1]
        else:
            prev = Sentence(sequence=[])
        if index2 < len(self._document.sentences):
            next = self._document.sentences[index2]
        else:
            next = Sentence(sequence=[])
        return {'main': self._document.sentences[item], 'prev': prev, 'next': next}


class MorphoDataset(Dataset):
    def __init__(self, document: Document):
        self._document = document

    def __len__(self):
        return len(self._document.sentences)

    def __getitem__(self, item):
        return self._document.sentences[item]


class LemmaDataset(Dataset):
    def __init__(self, document: Document, for_training=True):
        self._examples = []
        lookup = {}
        for sent in document.sentences:
            lang_id = sent.lang_id
            for w in sent.words:
                word = w.word
                lemma = w.lemma
                upos = w.upos

                key = (word, lang_id, upos)
                if key not in lookup or for_training is False:
                    lookup[key] = 1
                    example = {'word': word, 'upos': upos, 'lang_id': lang_id, 'target': lemma}
                    self._examples.append(example)

    def __len__(self):
        return len(self._examples)

    def __getitem__(self, item):
        return self._examples[item]


class CompoundDataset(Dataset):
    def __init__(self, document: Document, for_training=True):
        self._examples = []
        lookup = {}
        for sent in document.sentences:
            lang_id = sent.lang_id
            if for_training is True:
                for t in sent.tokens:
                    if len(t.words) > 1:
                        word = t.text
                        target = ' '.join([w.word for w in t.words])
                        key = (word, lang_id)
                        # if key not in lookup:
                        lookup[key] = 1
                        example = {'word': word, 'lang_id': lang_id, 'target': target}
                        self._examples.append(example)
            else:
                for t in sent.tokens:
                    example = {'word': t.words[0].word, 'lang_id': lang_id, 'target': ""}
                    self._examples.append(example)

    def __len__(self):
        return len(self._examples)

    def __getitem__(self, item):
        return self._examples[item]


class Word2TargetCollate:
    def __init__(self, encodings: Encodings):
        self._encodings = encodings
        self._start_index = len(encodings.char2int)
        self._stop_index = len(encodings.char2int) + 1
        self._eol = len(encodings.char2int)

    def collate_fn(self, batch):
        max_input_len = max([len(e['word']) for e in batch])
        max_target_len = max([len(e['target']) for e in batch])
        n = len(batch)
        x_char = np.zeros((n, max_input_len + 2), dtype=np.long)
        x_case = np.zeros((n, max_input_len + 2), dtype=np.long)
        y_char = np.zeros((n, max_target_len + 1), dtype=np.long)
        y_case = np.zeros((n, max_target_len + 1), dtype=np.long)
        x_lang = np.zeros(n, dtype=np.long)
        x_upos = np.zeros(n, dtype=np.long)

        for ii in range(n):
            word = batch[ii]['word']
            target = batch[ii]['target']
            lang_id = batch[ii]['lang_id']
            upos = ''
            if 'upos' in batch[ii]:
                upos = batch[ii]['upos']
            x_lang[ii] = lang_id + 1
            if upos in self._encodings.upos2int:
                x_upos[ii] = self._encodings.upos2int[upos]
            else:
                x_upos[ii] = self._encodings.upos2int['<UNK>']

            i_char = 1
            x_char[ii, 0] = self._start_index
            for ch in word:
                if ch.lower() == ch.upper():
                    x_case[ii, i_char] = 1  # symbol
                elif ch.lower() != ch:
                    x_case[ii, i_char] = 2  # uppercase
                else:
                    x_case[ii, i_char] = 3  # lowercase
                if ch.lower() in self._encodings.char2int:
                    x_char[ii, i_char] = self._encodings.char2int[ch.lower()]
                else:
                    x_char[ii, i_char] = self._encodings.char2int['<UNK>']
                i_char += 1
            x_char[ii, i_char] = self._stop_index

            i_char = 0
            for ch in target:
                if ch.lower() == ch.upper():
                    y_case[ii, i_char] = 1  # symbol
                elif ch.lower() != ch:
                    y_case[ii, i_char] = 2  # uppercase
                else:
                    y_case[ii, i_char] = 3  # lowercase
                if ch.lower() in self._encodings.char2int:
                    y_char[ii, i_char] = self._encodings.char2int[ch.lower()]
                else:
                    y_char[ii, i_char] = self._encodings.char2int['<UNK>']
                i_char += 1
            y_char[ii, i_char] = self._eol

        rez = {
            'x_char': torch.tensor(x_char),
            'x_case': torch.tensor(x_case),
            'x_lang': torch.tensor(x_lang),
            'x_upos': torch.tensor(x_upos),
            'y_char': torch.tensor(y_char),
            'y_case': torch.tensor(y_case)
        }
        return rez


class MorphoCollate:
    def __init__(self, encodings: Encodings, add_parsing=False, rhl_win_size=7):
        self._encodings = encodings
        self._add_parsing = add_parsing
        self._rhl_win_size = rhl_win_size

    def collate_fn(self, batch: [Sentence]):
        a_sent_len = [len(sent.words) for sent in batch]
        a_word_len = []

        x_word_embeddings = [[] for _ in range(len(batch[0].words[0].emb))]
        for sent in batch:
            for word in sent.words:
                a_word_len.append(len(word.word))
                for ii in range(len(word.emb)):
                    x_word_embeddings[ii].append(word.emb[ii])
        x_sent_len = np.array(a_sent_len, dtype=np.long)
        x_word_len = np.array(a_word_len, dtype=np.long)
        max_sent_len = np.max(x_sent_len)
        max_word_len = np.max(x_word_len)
        x_sent_masks = np.zeros((len(batch), max_sent_len), dtype=np.float)
        x_word_masks = np.zeros((x_word_len.shape[0], max_word_len), dtype=np.float)

        x_sent = np.zeros((len(batch), max_sent_len), dtype=np.long)
        x_word = np.zeros((x_word_len.shape[0], max_word_len), dtype=np.long)
        x_word_case = np.zeros((x_word_len.shape[0], max_word_len), dtype=np.long)
        c_word = 0
        x_lang_sent = np.zeros((len(batch)), dtype=np.long)
        x_lang_word = []

        y_upos = np.zeros((x_sent.shape[0], x_sent.shape[1]), dtype=np.long)
        y_xpos = np.zeros((x_sent.shape[0], x_sent.shape[1]), dtype=np.long)
        y_attrs = np.zeros((x_sent.shape[0], x_sent.shape[1]), dtype=np.long)

        y_head = np.zeros((x_sent.shape[0], x_sent.shape[1]), dtype=np.long)
        y_label = np.zeros((x_sent.shape[0], x_sent.shape[1]), dtype=np.long)
        y_rhl = np.zeros((x_sent.shape[0], x_sent.shape[1]), dtype=np.long)

        for iSent in range(len(batch)):
            sent = batch[iSent]
            x_lang_sent[iSent] = sent.lang_id + 1
            for iWord in range(len(sent.words)):
                word = sent.words[iWord]
                y_head[iSent, iWord] = word.head
                rhl = word.head - iWord + self._rhl_win_size
                rhl = np.clip(rhl, 0, self._rhl_win_size * 2 - 1)
                y_rhl[iSent, iWord] = rhl

                if word.label in self._encodings.label2int:
                    y_label[iSent, iWord] = self._encodings.label2int[word.label]
                if word.upos in self._encodings.upos2int:
                    y_upos[iSent, iWord] = self._encodings.upos2int[word.upos]
                if word.xpos in self._encodings.xpos2int:
                    y_xpos[iSent, iWord] = self._encodings.xpos2int[word.xpos]
                if word.attrs in self._encodings.attrs2int:
                    y_attrs[iSent, iWord] = self._encodings.attrs2int[word.attrs]
                word = sent.words[iWord].word
                x_sent_masks[iSent, iWord] = 1
                w = word.lower()
                x_lang_word.append(sent.lang_id + 1)
                if w in self._encodings.word2int:
                    x_sent[iSent, iWord] = self._encodings.word2int[w]
                else:
                    x_sent[iSent, iWord] = self._encodings.word2int['<UNK>']

                for iChar in range(len(word)):
                    x_word_masks[c_word, iChar] = 1
                    ch = word[iChar]
                    if ch.lower() == ch.upper():  # symbol
                        x_word_case[c_word, iChar] = 1
                    elif ch.lower() != ch:  # upper
                        x_word_case[c_word, iChar] = 2
                    else:  # lower
                        x_word_case[c_word, iChar] = 3
                    ch = ch.lower()
                    if ch in self._encodings.char2int:
                        x_word[c_word, iChar] = self._encodings.char2int[ch]
                    else:
                        x_word[c_word, iChar] = self._encodings.char2int['<UNK>']
                c_word += 1

        x_lang_word = np.array(x_lang_word)
        x_word_embeddings = [torch.tensor(t) for t in x_word_embeddings]
        response = {
            'x_sent': torch.tensor(x_sent),
            'x_lang_sent': torch.tensor(x_lang_sent),
            'x_word': torch.tensor(x_word),
            'x_word_case': torch.tensor(x_word_case),
            'x_lang_word': torch.tensor(x_lang_word),
            'x_sent_len': torch.tensor(x_sent_len),
            'x_word_len': torch.tensor(x_word_len),
            'x_sent_masks': torch.tensor(x_sent_masks),
            'x_word_masks': torch.tensor(x_word_masks),
            'x_word_embeddings': x_word_embeddings,
            'y_upos': torch.tensor(y_upos),
            'y_xpos': torch.tensor(y_xpos),
            'y_attrs': torch.tensor(y_attrs)
        }

        if self._add_parsing:
            response['y_head'] = torch.tensor(y_head)
            response['y_label'] = torch.tensor(y_label)
            response['y_rhl'] = torch.tensor(y_rhl)

        return response


Arc = namedtuple('Arc', ('tail', 'weight', 'head'))


class GreedyDecoder:
    def _valid(self, arc, tree):
        # just one head
        for sa in tree:
            if sa.tail == arc.tail:
                return False
        stack = [arc.head]
        pos = 0
        used = [False] * len(tree)
        while pos < len(stack):
            for zz in range(len(tree)):
                if tree[zz].tail == stack[pos] and not used[zz]:
                    used[zz] = True
                    stack.append(tree[zz].head)
                    if tree[zz].head == arc.tail:
                        return False
            pos += 1
            # print pos,len(stack)
        return True

    def _get_sort_key(self, item):
        return item.weight

    def _greedy_tree(self, arcs):
        arcs = sorted(arcs, key=self._get_sort_key, reverse=True)
        # print arcs
        final_tree = []
        for index in range(len(arcs)):
            if self._valid(arcs[index], final_tree):
                final_tree.append(arcs[index])
                # print arcs[index]
        return final_tree

    def _make_ordered_list(self, tree, nWords):
        lst = [0] * nWords  # np.zeros(nWords)
        for arc in tree:
            # arc = tree[index]
            tail = arc.tail
            head = arc.head
            lst[tail] = head
        return lst[1:]

    def decode(self, score, lens):
        best_tree_list = []
        for ii in range(score.shape[0]):
            # norm_score = score[ii, :lens[ii], :lens[ii]]
            norm_score = np.zeros((lens[ii] + 1, lens[ii] + 1))
            for wii in range(lens[ii]):
                for wjj in range(lens[ii] + 1):
                    norm_score[wii + 1, wjj] = score[ii, wii, wjj]
            nWords = norm_score.shape[0]  # len(norm_score)
            g = []
            for iSrc in range(1, nWords):
                for iDst in range(1, nWords):
                    if iDst != iSrc:
                        a = Arc(iSrc, norm_score[iSrc][iDst], iDst)
                        g.append(a)
            tree = self._greedy_tree(g)
            best_tree = self._make_ordered_list(tree, nWords)
            best_tree_list.append(best_tree)
        return best_tree_list


# Code adapted from https://github.com/tdozat/Parser-v3/blob/master/scripts/chuliu_edmonds.py
class ChuLiuEdmondsDecoder:
    def __init__(self):
        pass

    def _tarjan(self, tree):
        indices = -np.ones_like(tree)
        lowlinks = -np.ones_like(tree)
        onstack = np.zeros_like(tree, dtype=bool)
        stack = list()
        _index = [0]
        cycles = []

        def strong_connect(i):
            _index[0] += 1
            index = _index[-1]
            indices[i] = lowlinks[i] = index - 1
            stack.append(i)
            onstack[i] = True
            dependents = np.where(np.equal(tree, i))[0]
            for j in dependents:
                if indices[j] == -1:
                    strong_connect(j)
                    lowlinks[i] = min(lowlinks[i], lowlinks[j])
                elif onstack[j]:
                    lowlinks[i] = min(lowlinks[i], indices[j])

            # There's a cycle!
            if lowlinks[i] == indices[i]:
                cycle = np.zeros_like(indices, dtype=bool)
                while stack[-1] != i:
                    j = stack.pop()
                    onstack[j] = False
                    cycle[j] = True
                stack.pop()
                onstack[i] = False
                cycle[i] = True
                if cycle.sum() > 1:
                    cycles.append(cycle)
            return

        for i in range(len(tree)):
            if indices[i] == -1:
                strong_connect(i)
        return cycles

    def _chuliu_edmonds(self, scores):
        scores *= (1 - np.eye(scores.shape[0]))
        scores[0] = 0
        scores[0, 0] = 1
        tree = np.argmax(scores, axis=1)
        cycles = self._tarjan(tree)
        if not cycles:
            return tree
        else:
            # t = len(tree); c = len(cycle); n = len(noncycle)
            # locations of cycle; (t) in [0,1]
            cycle = cycles.pop()
            # indices of cycle in original tree; (c) in t
            cycle_locs = np.where(cycle)[0]
            # heads of cycle in original tree; (c) in t
            cycle_subtree = tree[cycle]
            # scores of cycle in original tree; (c) in R
            cycle_scores = scores[cycle, cycle_subtree]
            # total score of cycle; () in R
            cycle_score = cycle_scores.prod()

            # locations of noncycle; (t) in [0,1]
            noncycle = np.logical_not(cycle)
            # indices of noncycle in original tree; (n) in t
            noncycle_locs = np.where(noncycle)[0]
            # print(cycle_locs, noncycle_locs)

            # scores of cycle's potential heads; (c x n) - (c) + () -> (n x c) in R
            metanode_head_scores = scores[cycle][:, noncycle] / cycle_scores[:, None] * cycle_score
            # scores of cycle's potential dependents; (n x c) in R
            metanode_dep_scores = scores[noncycle][:, cycle]
            # best noncycle head for each cycle dependent; (n) in c
            metanode_heads = np.argmax(metanode_head_scores, axis=0)
            # best cycle head for each noncycle dependent; (n) in c
            metanode_deps = np.argmax(metanode_dep_scores, axis=1)

            # scores of noncycle graph; (n x n) in R
            subscores = scores[noncycle][:, noncycle]
            # pad to contracted graph; (n+1 x n+1) in R
            subscores = np.pad(subscores, ((0, 1), (0, 1)), 'constant')
            # set the contracted graph scores of cycle's potential heads; (c x n)[:, (n) in n] in R -> (n) in R
            subscores[-1, :-1] = metanode_head_scores[metanode_heads, np.arange(len(noncycle_locs))]
            # set the contracted graph scores of cycle's potential dependents; (n x c)[(n) in n] in R-> (n) in R
            subscores[:-1, -1] = metanode_dep_scores[np.arange(len(noncycle_locs)), metanode_deps]

            # MST with contraction; (n+1) in n+1
            contracted_tree = self._chuliu_edmonds(subscores)
            # head of the cycle; () in n
            # print(contracted_tree)
            cycle_head = contracted_tree[-1]
            # fixed tree: (n) in n+1
            contracted_tree = contracted_tree[:-1]
            # initialize new tree; (t) in 0
            new_tree = -np.ones_like(tree)
            # print(0, new_tree)
            # fixed tree with no heads coming from the cycle: (n) in [0,1]
            contracted_subtree = contracted_tree < len(contracted_tree)
            # add the nodes to the new tree (t)[(n)[(n) in [0,1]] in t] in t = (n)[(n)[(n) in [0,1]] in n] in t
            new_tree[noncycle_locs[contracted_subtree]] = noncycle_locs[contracted_tree[contracted_subtree]]
            # print(1, new_tree)
            # fixed tree with heads coming from the cycle: (n) in [0,1]
            contracted_subtree = np.logical_not(contracted_subtree)
            # add the nodes to the tree (t)[(n)[(n) in [0,1]] in t] in t = (c)[(n)[(n) in [0,1]] in c] in t
            new_tree[noncycle_locs[contracted_subtree]] = cycle_locs[metanode_deps[contracted_subtree]]
            # print(2, new_tree)
            # add the old cycle to the tree; (t)[(c) in t] in t = (t)[(c) in t] in t
            new_tree[cycle_locs] = tree[cycle_locs]
            # print(3, new_tree)
            # root of the cycle; (n)[() in n] in c = () in c
            cycle_root = metanode_heads[cycle_head]
            # add the root of the cycle to the new tree; (t)[(c)[() in c] in t] = (c)[() in c]
            new_tree[cycle_locs[cycle_root]] = noncycle_locs[cycle_head]
            # print(4, new_tree)
            return new_tree

    def _chuliu_edmonds_one_root(self, scores):
        """"""

        scores = scores.astype(np.float64)
        tree = self._chuliu_edmonds(scores)
        roots_to_try = np.where(np.equal(tree[1:], 0))[0] + 1
        if len(roots_to_try) == 1:
            return tree

        # Look at all roots that are more likely than we would expect
        if len(roots_to_try) == 0:
            roots_to_try = np.where(scores[1:, 0] >= 1 / len(scores))[0] + 1
        # *sigh* just grab the most likely one
        if len(roots_to_try) == 0:
            roots_to_try = np.array([np.argmax(scores[1:, 0]) + 1])

        # -------------------------------------------------------------
        def set_root(scores, root):
            root_score = scores[root, 0]
            scores = np.array(scores)
            scores[1:, 0] = 0
            scores[root] = 0
            scores[root, 0] = 1
            return scores, root_score

        # -------------------------------------------------------------

        best_score, best_tree = -np.inf, None  # This is what's causing it to crash
        for root in roots_to_try:
            _scores, root_score = set_root(scores, root)
            _tree = self._chuliu_edmonds(_scores)
            tree_probs = _scores[np.arange(len(_scores)), _tree]
            tree_score = np.log(tree_probs).sum() + np.log(root_score) if tree_probs.all() else -np.inf
            if tree_score > best_score:
                best_score = tree_score
                best_tree = _tree
        try:
            assert best_tree is not None
        except:
            with open('debug.log', 'w') as f:
                f.write('{}: {}, {}\n'.format(tree, scores, roots_to_try))
                f.write('{}: {}, {}, {}\n'.format(_tree, _scores, tree_probs, tree_score))
            raise
        return best_tree

    def decode(self, score, lens):
        best_tree_list = []
        for ii in range(score.shape[0]):
            # norm_score = score[ii, :lens[ii], :lens[ii]]
            norm_score = np.zeros((lens[ii] + 1, lens[ii] + 1))
            for wii in range(lens[ii]):
                for wjj in range(lens[ii] + 1):
                    norm_score[wii + 1, wjj] = score[ii, wii, wjj]
            nWords = norm_score.shape[0]  # len(norm_score)
            norm_score *= (1 - np.eye(nWords))
            tree = self._chuliu_edmonds_one_root(norm_score)
            best_tree_list.append(tree[1:])
        return best_tree_list



#
# Author: Tiberiu Boros
#
# Copyright (c) 2018 Adobe Systems Incorporated. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import dynet as dy
import numpy as np


class CompoundWordExpander:
    def __init__(self, config, encodings, embeddings, runtime=False):
        self.config = config
        self.word_embeddings = embeddings
        self.encodings = encodings
        self.model = dy.Model()
        self.trainer = dy.AdamTrainer(self.model, alpha=2e-3, beta_1=0.9, beta_2=0.9)

        from cube.character_embeddings import CharacterNetwork
        self.encoder = CharacterNetwork(self.config.character_embeddings_size, encodings, self.config.encoder_size,
                                        self.config.encoder_layers, self.config.character_embeddings_size, self.model,
                                        runtime=runtime)

        self.decoder = dy.VanillaLSTMBuilder(self.config.decoder_layers, self.config.encoder_size * 2,
                                             self.config.decoder_size, self.model)
        self.decoder_start_lookup = self.model.add_lookup_parameters((1, self.config.encoder_size * 2))

        # self.att_w1 = self.model.add_parameters(
        #     (self.config.character_embeddings_size * 2, self.config.encoder_size * 2))
        # self.att_w2 = self.model.add_parameters(
        #     (self.config.character_embeddings_size * 2, self.config.decoder_size))
        # self.att_v = self.model.add_parameters((1, self.config.character_embeddings_size * 2))

        self.softmax_w = self.model.add_parameters(
            (len(self.encodings.char2int) + 4,
             self.config.decoder_size))  # all known characters except digits with COPY, INC, TOK and EOS
        self.softmax_b = self.model.add_parameters((len(self.encodings.char2int) + 4))

        self.softmax_comp_w = self.model.add_parameters((2, self.config.character_embeddings_size))
        self.softmax_comp_b = self.model.add_parameters((2))

        self.label2int = {}
        ofs = len(self.encodings.char2int)
        self.label2int['<EOS>'] = ofs
        self.label2int['<TOK>'] = ofs + 1
        self.label2int['<COPY>'] = ofs + 2
        self.label2int['<INC>'] = ofs + 3

        self.losses = []

    def start_batch(self):
        self.losses = []
        dy.renew_cg()

    def end_batch(self):
        total_loss = 0
        if len(self.losses) != 0:
            loss = dy.esum(self.losses)
            self.losses = []
            total_loss = loss.value()
            loss.backward()
            self.trainer.update()
        dy.renew_cg()
        return total_loss

    def learn(self, seq):
        losses = []
        examples = self._get_examples(seq)

        for example in examples:
            y_pred, encoder_states = self._predict_is_compound_entry(example.source, runtime=False)
            if not example.should_expand:
                losses.append(-dy.log(dy.pick(y_pred, 0)))
            else:
                losses.append(-dy.log(dy.pick(y_pred, 1)))
                losses.append(self._learn_transduction(example.source, example.destination, encoder_states))
        loss = dy.esum(losses)
        self.losses.append(loss)

    def _compute_transduction_states(self, source, destination):
        a = np.zeros((len(source) + 1, len(destination) + 1))
        for i in range(len(source) + 1):
            a[i, 0] = i

        for i in range(len(destination) + 1):
            a[0, i] = i

        for i in range(1, len(source) + 1):
            for j in range(1, len(destination) + 1):
                cost = 0
                if source[i - 1] != destination[j - 1]:
                    cost = 1
                m = min([a[i - 1, j - 1], a[i - 1, j], a[i, j - 1]])
                a[i, j] = m + cost

        alignments = [-1] * len(destination)

        i = len(source)
        j = len(destination)
        while i > 1 or j > 1:
            if source[i - 1] == destination[j - 1]:
                alignments[j - 1] = i - 1
            if i == 1:
                j -= 1
            elif j == 1:
                i -= 1
            else:
                if a[i - 1, j - 1] <= a[i - 1, j] and a[i - 1, j - 1] <= a[i, j - 1]:
                    i -= 1
                    j -= 1
                elif a[i - 1][j] <= a[i - 1, j - 1] and a[i - 1, j] <= a[i, j - 1]:
                    i -= 1
                else:
                    j -= 1
        if source[i - 1] == destination[j - 1]:
            alignments[j - 1] = i - 1

        y_pred = []
        index_src = 0
        index_dst = 0
        while index_dst < len(destination):
            if alignments[index_dst] == index_src:
                y_pred.append("<COPY>")
                index_dst += 1
            elif alignments[index_dst] == -1:
                if destination[index_dst] == "\t":
                    y_pred.append("<TOK>")
                    index_dst += 1
                else:
                    y_pred.append(destination[index_dst])
                    index_dst += 1
            else:
                y_pred.append("<INC>")
                index_src += 1

        y_pred.append("<EOS>")
        return y_pred

    def _attend(self, input_vectors, state):
        w1 = self.att_w1.expr(update=True)
        w2 = self.att_w2.expr(update=True)
        v = self.att_v.expr(update=True)
        attention_weights = []

        w2dt = w2 * state.h()[-1]
        for input_vector in input_vectors:
            attention_weight = v * dy.tanh(w1 * input_vector + w2dt)
            attention_weights.append(attention_weight)

        attention_weights = dy.softmax(dy.concatenate(attention_weights))

        output_vectors = dy.esum(
            [vector * attention_weight for vector, attention_weight in zip(input_vectors, attention_weights)])

        return output_vectors

    def _decode(self, encoder_states, runtime=True, max_preds=-1, gs_labels=None):
        y_pred = []
        num_preds = 0
        lstm = self.decoder.initial_state().add_input(self.decoder_start_lookup[0])
        i_src = 0
        while num_preds < max_preds:
            # input = self._attend(encoder_states, lstm)
            input = encoder_states[i_src]
            lstm = lstm.add_input(input)
            softmax_out = dy.softmax(self.softmax_w.expr(update=True) * lstm.output() + self.softmax_b.expr(update=True))
            y_pred.append(softmax_out)

            if runtime:
                if np.argmax(softmax_out.npvalue()) == self.label2int['<EOS>']:
                    return y_pred
                elif np.argmax(softmax_out.npvalue()) == self.label2int['<INC>'] and i_src < len(encoder_states) - 1:
                    i_src += 1
            else:
                if gs_labels[num_preds] == '<INC>' and i_src < len(encoder_states) - 1:
                    i_src += 1
            num_preds += 1
        return y_pred

    def _learn_transduction(self, source, destination, encoder_states):
        losses = []
        y_target = self._compute_transduction_states(source, destination)
        y_predicted = self._decode(encoder_states, runtime=False, max_preds=len(y_target), gs_labels=y_target)
        for y_real, y_pred in zip(y_target, y_predicted):
            if y_real in self.label2int:
                losses.append(-dy.log(dy.pick(y_pred, self.label2int[y_real])))
            else:
                if y_real in self.encodings.char2int:
                    losses.append(-dy.log(dy.pick(y_pred, self.encodings.char2int[y_real])))
                # else:
                #    print source + "\t\t" + destination

        return dy.esum(losses)

    def _predict_is_compound_entry(self, word, runtime=True):
        emb, states = self.encoder.compute_embeddings(word, runtime=runtime)
        output = dy.softmax(self.softmax_comp_w.expr(update=True) * emb + self.softmax_comp_b.expr(update=True))
        return output, states

    def _transduce(self, source, encoder_states):
        tokens = []
        y_pred = self._decode(encoder_states, runtime=True, max_preds=100)

        i_src = 0
        token = ""
        for y in y_pred:
            y = np.argmax(y.npvalue())
            if y == self.label2int['<INC>'] and i_src < len(encoder_states) - 1:
                i_src += 1
            elif y == self.label2int['<COPY>']:
                if i_src < len(source):
                    token += source[i_src]
            elif y == self.label2int['<TOK>'] or y == self.label2int['<EOS>']:
                tokens.append(token)
                token = ""
            else:
                if y < len(self.encodings.characters):
                    token += self.encodings.characters[y]

        return tokens

    def tag_token(self, word):
        dy.renew_cg()
        compound = False
        word = unicode(word, 'utf-8')
        tokens = []
        ce_out, encoder_states = self._predict_is_compound_entry(word, runtime=True)
        if np.argmax(ce_out.npvalue()) == 1:
            tokens = self._transduce(word, encoder_states)
            compound = True
        return compound, tokens

    def tag(self, seq):
        dy.renew_cg()
        new_seq = []
        index = 1
        for entry in seq:
            if not entry.is_compound_entry:
                ce_out, encoder_states = self._predict_is_compound_entry(unicode(entry.word, 'utf-8'), runtime=True)
                if np.argmax(ce_out.npvalue()) == 0:
                    entry.index = index
                    new_seq.append(entry)
                    index += 1
                else:
                    compounds = self._transduce(unicode(entry.word, 'utf-8'), encoder_states)
                    # bug because _transduce may return empty tokens
                    valid_tokens = []
                    for token in compounds:
                        if token.strip() != "":
                            valid_tokens.append(token)
                    compounds = valid_tokens
                    if len(compounds) <= 1:
                        entry.index = index
                        new_seq.append(entry)
                        index += 1
                    else:
                        entry.index = str(index) + '-' + str(index + len(compounds) - 1)
                        entry.is_compound_entry = True
                        entry.upos='_'
                        entry.xpos = '_'
                        entry.attrs = '_'
                        entry.label = '_'
                        entry.head = '_'
                        entry.deps = '_'
                        new_seq.append(entry)
                        for word in compounds:
                            from io_utils.conll import ConllEntry
                            entry = ConllEntry(index, word.encode('utf-8'), word.encode('utf-8'), '_', '_', '_', '0',
                                               '_',
                                               '_',
                                               '')
                            new_seq.append(entry)
                            index += 1

        return new_seq

    def _get_examples(self, seq):
        examples = []
        cww = 0
        for entry in seq:
            if cww == 0:
                et = ExpandedToken(source=unicode(entry.word, 'utf-8'))
                if entry.is_compound_entry:
                    et.should_expand = True
                    et.destination = u''
                    interval = entry.index
                    interval = interval.split("-")
                    stop = int(interval[1])
                    start = int(interval[0])
                    cww = stop - start + 1
                else:
                    et.destination = et.source
                    examples.append(et)
            else:
                et.destination += "\t" + unicode(entry.word, 'utf-8')
                cww -= 1
                if cww == 0:
                    et.destination = et.destination.strip()
                    examples.append(et)

        return examples

    def save(self, filename):
        self.model.save(filename)

    def load(self, path):
        self.model.populate(path)

    def expand_sequences(self, sequences):
        new_sequences = []
        for sequence in sequences:
            new_sequence = self.tag(sequence)
            new_sequences.append(new_sequence)
        return new_sequences


class ExpandedToken:
    def __init__(self, source=None, destination=None, should_expand=False):
        self.source = source
        self.destination = destination
        self.should_expand = should_expand



""" this needs to be deleted """

import json, os, sys
# Append parent dir to sys path.
from shutil import rmtree, copyfile
import logging
from tqdm.autonotebook import tqdm as tqdm

parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
os.sys.path.insert(0, parent_dir)

from cube.io_utils.modelstore import ModelStore

def check (filename, language_family, files):
    for file in files:
        if filename in file:
            return 1
    print(f"{language_family} is missing [{filename}].\n")
    with open("log.txt", "a") as f:
        f.write(f"{language_family} is missing [{filename}].\n")
    return 0

if __name__ == "__main__":
    refresh = False
    folder_with_base_confs = os.path.abspath("scripts//train//2.7//language//")
    folder_with_all_trained_models = "/media/echo/5CA436CBA436A802/work/models" #os.path.abspath("models")
    folder_where_to_output_everything = "/media/echo/5CA436CBA436A802/work/nlp-cube-models" #os.path.abspath("nlp-cube-models")
    url_root_for_models = "https://github.com/adobe/NLP-Cube-Models/blob/3.0/models/" # !! make sure it ends with an /

    logger = logging.getLogger("cube")
    log_handler = logging.StreamHandler()
    log_formatter = logging.Formatter(
        fmt="[%(levelname)8s | %(asctime)s | %(filename)-20s:%(lineno)3s | %(funcName)-26s] %(message)s",
        datefmt='%Y-%m-%d %H:%M:%S')
    log_handler.setFormatter(log_formatter)
    logger.addHandler(log_handler)
    logger.setLevel(logging.DEBUG)


    """
        Read base confs and get a list of language families
    """
    language_family_confs = os.listdir(folder_with_base_confs) # this just lists the files without path
    language_family_confs = [os.path.abspath(os.path.join(folder_with_base_confs, x)) for x in language_family_confs]
    language_family_confs = [x for x in language_family_confs if os.path.isfile(x) and x.endswith(".yaml")]
    print("I see {} language families in {}.".format(len(language_family_confs), folder_with_base_confs))

    d = {}
    for f in language_family_confs:
        js = json.load(open(f, "r"))
        # list of lists where [0] is the lang code
        _, name = os.path.split(f)
        if name == "all.json":
            continue
        d[name.replace(".json","")] = set()

        for e in js:
            p = e[0]
            d[name.replace(".json", "")].add(p)
            if "_" in p:
                d[name.replace(".json", "")].add(p.split("_")[0])
        if len(d[name.replace(".json","")]) == 0:
            del d[name.replace(".json","")]
        else:
            print("\tlanguage family {} has {} codes: {}".format(name.replace(".json",""), len(d[name.replace(".json","")]), d[name.replace(".json","")]))

    """    
        For each language family, create a folder with the packed and split model in output folder
        Mark language family as valid 
    """
    import uuid

    catalog = {}
    catalog["default"] = []

    os.makedirs(folder_where_to_output_everything, exist_ok=True)
    for fam in d:
        print("Processing language family {}:".format(fam))

        # copy all files in a temp folder
        model_files = os.listdir(folder_with_all_trained_models)
        #print(model_files)
        model_files = [x for x in model_files if x.startswith(fam+"-") and ".last" not in x and ".zip" not in x]

        # check model is complete
        list_of_files = [
            "tokenizer.config",
            "tokenizer.encodings",
            "tokenizer.yaml",

        ]
        #check("")

        #print(model_files)
        model_files = [os.path.abspath(os.path.join(folder_with_all_trained_models,x)) for x in model_files]
        if len(model_files) == 0:
            print("\t no model files found, skipping ...")
            continue
        print("\t found {} model files, copying to temp folder ...".format(len(model_files)))

        temp_folder = os.path.join(folder_where_to_output_everything, str(uuid.uuid4().hex))
        os.mkdir(temp_folder)
        for src_file in tqdm(model_files):
            _, name = os.path.split(src_file)
            dst_file = os.path.join(temp_folder, name)
            #print((src_file, dst_file))
            copyfile(src_file, dst_file)

        # pack folder in zip file
        print("\t packing model ... ")
        zip_file_path = os.path.join(temp_folder, fam+".zip")
        ModelStore._pack_model(input_folder = temp_folder, output_file_path = zip_file_path)

        # split for upload to github
        print("\t splitting model zip ...")
        output_folder = os.path.join(folder_where_to_output_everything, fam)
        os.makedirs(output_folder, exist_ok=True)
        ModelStore._split_packed_model(file_path = zip_file_path, output_folder = output_folder)

        # delete temp folder
        print("\t deleting temp folder ...")
        if os.path.exists(temp_folder):
            rmtree(temp_folder, ignore_errors=True)

        # make a catalog entry
        entry = {}
        entry["languages"] = list(d[fam])
        entry["model"] = url_root_for_models + fam
        catalog["default"].append(entry)


    """    
        Generate catalog
    """
    print("Writing catalog ...")
    json.dump(catalog, open(os.path.join(folder_where_to_output_everything, "catalog.json"), "w", encoding="utf8"), indent=4, sort_keys=True)

    print("Done.")



import setuptools
def parse_requirements(filename, session=None):
    """ load requirements from a pip requirements file """
    lineiter = (line.strip() for line in open(filename))
    return [line for line in lineiter if line and not line.startswith("#")]

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="nlpcube",
    version="0.3.1.1",
    author="Multiple authors",
    author_email="tiberiu44@gmail.com",
    description="Natural Language Processing Toolkit with support for tokenization, sentence splitting, lemmatization, tagging and parsing for more than 60 languages",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/adobe/NLP-Cube",
    packages=setuptools.find_packages(),
    install_requires = parse_requirements('requirements.txt', session=False),
    classifiers=(
        "Programming Language :: Python :: 3.0",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ),
)



import sys, os, yaml
from pprint import pprint
from bs4 import BeautifulSoup

path_to_corpus_folder = "corpus/ud-treebanks-v2.5"
path_to_save_folder = "scripts/train/2.5"

path_to_corpus_folder = "corpus/ud-treebanks-v2.7/"
path_to_save_folder = "scripts/train/2.7"

# read default language map
default_language_map = {}
with open("scripts/default_language.txt","r") as f:
    lines = f.readlines()
    for l in lines:
        p = l.strip().split("\t")
        major, treebank = p[0], p[0]+"_"+p[1]
        default_language_map[major] = treebank

# read UD extracted table file as html <-- copy-paste from UD's website of the table only
with open("scripts/ud_table.html","r", encoding="utf8") as f:
    source = f.read()

soup = BeautifulSoup(source, features="html.parser")
table = []

##### STEP 1, extract data from html table
header = True
for tag in soup.div.findChildren(recursive=False):
    if str(tag).strip() == "" or len(str(tag))<50:
        continue
    if header:
        #print(tag)
        for elem in tag.find_all("span", class_="doublewidespan", recursive=False):
            if elem.text.strip() != "":
                language_name = elem.text
        print("Language name: "+ language_name)

        for elem in tag.find_all("span", class_="triplewidespan", recursive=False):
            if elem.text.strip() != "":
                language_family = [x.strip() for x in elem.text.split(",")]
        print("Language family: {}".format(language_family))

        header = False
    else:
        cheader = True
        for content in tag.find_all("div", class_="ui-accordion-icons", recursive=False): # just 1 content
            for content_div in content.findChildren(recursive=False): # list of divs
                #print(content_div)
                #print("_" * 50)

                if cheader: # get name of treebank
                    for sp in content_div.find_all("span", class_="doublewidespan", recursive=False):
                        #print(sp)
                        if sp.text.strip() != "":
                            treebank_name = sp.text
                    print("Treebank name: "+treebank_name)
                    cheader = False
                else: # get folder name
                    for sp in content_div.find_all("a", href=True):
                        #print(sp["href"])
                        link = str(sp["href"]).strip()
                        if "UniversalDependencies" in link:
                            link = link[41:]
                            treebank_folder =  link[:link.find("/")]
                        elif "treebanks/" in link:
                            link = link[10:]
                            link = link[:link.find("/")]
                            language_code = link
                            major_language_code = link.split("_")[0]

                    print("Treebank folder: " + treebank_folder)
                    print("Language code: " + language_code)
                    print("Major language code: " + major_language_code)

                    if language_name == None:
                        raise Exception("Language name is none")
                    if language_family == None:
                        raise Exception("Language family none")
                    if treebank_name == None:
                        raise Exception("Treebank name is none")
                    if treebank_folder == None:
                        raise Exception("Treebank folder is none")

                    telem = {}
                    telem["language_name"] = language_name
                    telem["language_family"] = language_family
                    telem["treebank_name"] = treebank_name
                    telem["treebank_folder"] = treebank_folder
                    telem["language_code"] = language_code
                    telem["major_language_code"] = major_language_code
                    print(telem)
                    if "simonero" in language_code:
                        continue
                    if "_pud" in language_code:
                        continue
                    table.append(telem)
                    treebank_name, treebank_folder, language_code, major_language_code = None, None, None, None

                    cheader = True

        # reset for next language
        language_name, language_family, treebank_name, treebank_folder = None, None, None, None

        header = True
    #input('\n\n----------\n\n')

##### STEP 2, add paths to train dev test
valid = []
for l in table:
    train_file = os.path.join(path_to_corpus_folder, l["treebank_folder"], l["language_code"] + "-ud-train.conllu")
    dev_file = os.path.join(path_to_corpus_folder, l["treebank_folder"], l["language_code"] + "-ud-dev.conllu")
    test_file = os.path.join(path_to_corpus_folder, l["treebank_folder"], l["language_code"] + "-ud-test.conllu")

    # for PUD treebanks
    if l["language_code"].endswith("pud"): # languages only with test files
        l["train_file"] = [l["language_code"],train_file]
        l["dev_file"] = None
        l["test_file"] = None
    else: # regular treebank
        if os.path.exists(train_file):
            l["train_file"] = [l["language_code"],train_file]
        else:
            l["train_file"] = None
        if os.path.exists(dev_file):
            l["dev_file"] = [l["language_code"],dev_file]
        else:
            l["dev_file"] = None
        if os.path.exists(test_file):
            if l["dev_file"] == None:
                l["dev_file"] = [l["language_code"],test_file]
                l["test_file"] = [l["language_code"],test_file]
            else:
                l["test_file"] = [l["language_code"],test_file]
        else:
            l["test_file"] = None

    if l["train_file"] is not None:
        valid.append(l)
    else:
        print(" Language skipped due to no training file: ")
        pprint(l)
        print("\n")

table = valid


#### STEP 3, generate train files

# generate single language files
print("\n\nRunning single treebank ...\n")
folder = os.path.join(path_to_save_folder, "treebank")
os.makedirs(folder, exist_ok=True)

for l in table:
    obj = {
        "language_map": {l["language_code"]:l["language_code"]},
        "language_codes": [l["language_code"]],
        "train_files": {},
        "dev_files": {},
        "test_files": {}
    }

    if l["train_file"] is not None:
        obj["train_files"][l["train_file"][0]] = l["train_file"][1]
    if l["dev_file"] is not None:
        obj["dev_files"][l["dev_file"][0]] = l["dev_file"][1]
    if l["test_file"] is not None:
        obj["test_files"][l["test_file"][0]] = l["test_file"][1]
    else:
        if l["dev_file"] is not None:
            obj["test_files"][l["dev_file"][0]] = l["dev_file"][1]

    obj["language_codes"] = list(set([obj["language_map"][key] for key in obj["language_map"]]))

    if len(obj["train_files"]) == 0 :
        print("\tLanguage {} has zero training files, skipping. ".format(l["language_code"]))
        continue
    if len(obj["dev_files"]) == 0 :
        print("\tLanguage {} has zero dev files, skipping. ".format(l["language_code"]))
        continue
    if len(obj["test_files"]) == 0:
        print("\tLanguage {} has zero test files, skipping. ".format(l["language_code"]))
        continue

    filename = l["language_code"]+".yaml"
    with open(os.path.join(folder, filename), 'w') as f:
        data = yaml.dump(obj, f, sort_keys=True)

# generate per major language files (e.g. en)
print("\n\nRunning per language treebank aggregation ...\n")
folder = os.path.join(path_to_save_folder, "language")
os.makedirs(folder, exist_ok=True)

groups = {}
for l in table:
    lc = l["major_language_code"]
    if lc not in groups:
        groups[lc] = []
    groups[lc].append(l)

for g in groups:
    if g not in default_language_map:
        #raise Exception("\t\tLanguage {} does not have default language code, skipping!".format(g))
        print("\tLanguage {} ({}) does not have default language code, skipping!".format(groups[g][0]["language_name"], g))
        continue

    obj = {
        "language_map": {},
        "language_codes": [],
        "train_files": {},
        "dev_files": {},
        "test_files": {}
    }
    for l in groups[g]:
        # add full language name, lowercased
        if l["language_name"].lower() not in obj["language_map"]:
            obj["language_map"][l["language_name"].lower()] = default_language_map[g]
        # add major language code
        if g not in obj["language_map"]:
            obj["language_map"][g] = default_language_map[g]
        # add treebank code
        obj["language_map"][l["language_code"]] = l["language_code"]

        if l["train_file"] is not None:
            obj["train_files"][l["train_file"][0]] = l["train_file"][1]
        if l["dev_file"] is not None:
            obj["dev_files"][l["dev_file"][0]] = l["dev_file"][1]
        if l["test_file"] is not None:
            obj["test_files"][l["test_file"][0]] = l["test_file"][1]

    obj["language_codes"] = list(set([obj["language_map"][key] for key in obj["language_map"]]))

    if len(obj["train_files"]) == 0 :
        print("\tLanguage {} ({}) has zero training files, skipping. ".format(l["language_name"], g))
        continue
    if len(obj["dev_files"]) == 0 :
        print("\tLanguage {} ({}) has zero dev files, skipping. ".format(l["language_name"], g))
        continue
    if len(obj["test_files"]) == 0:
        print("\tLanguage {} ({}) has zero test files, skipping. ".format(l["language_name"], g))
        continue

    filename = l["major_language_code"]+".yaml"
    with open(os.path.join(folder, filename), 'w') as f:
        data = yaml.dump(obj, f, sort_keys=True)

# generate per language family
print("\n\nRunning per language family aggregation ...\n")
folder = os.path.join(path_to_save_folder, "family")
os.makedirs(folder, exist_ok=True)

groups = {}
for l in table:
    if l["major_language_code"] not in default_language_map:
        #raise Exception("\t\tLanguage {} does not have default language code, skipping!".format(g))
        print("\tLanguage {} ({}) does not have default language code, skipping!".format(l["language_name"], l["major_language_code"]))
        continue

    families = l["language_family"]
    for f in families:
        if f != "IE" and "Afro-Asiatic" not in f:
            if f not in groups:
                groups[f] = []

            groups[f].append(l)

for g in groups:
    obj = {
        "language_map": {},
        "language_codes": [],
        "train_files": {},
        "dev_files": {},
        "test_files": {}
    }

    for l in groups[g]:
        # add full language name, lowercased
        if l["language_name"].lower() not in obj["language_map"]:
            obj["language_map"][l["language_name"].lower()] = default_language_map[l["major_language_code"]]
        # add major language code
        if g not in obj["language_map"]:
            obj["language_map"][l["major_language_code"]] = default_language_map[l["major_language_code"]]
        # add treebank code
        obj["language_map"][l["language_code"]] = l["language_code"]

        if l["train_file"] is not None:
            obj["train_files"][l["train_file"][0]] = l["train_file"][1]
        if l["dev_file"] is not None:
            obj["dev_files"][l["dev_file"][0]] = l["dev_file"][1]
        if l["test_file"] is not None:
            obj["test_files"][l["test_file"][0]] = l["test_file"][1]

    obj["language_codes"] = list(set([obj["language_map"][key] for key in obj["language_map"]]))

    if len(obj["train_files"]) == 0 :
        print("\tLanguage {} (fam: {}) has zero training files, skipping. ".format(l["language_name"], g))
        pprint(obj)
        continue
    if len(obj["dev_files"]) == 0 :
        print("\tLanguage {} (fam: {}) has zero dev files, skipping. ".format(l["language_name"], g))
        pprint(obj)
        continue
    if len(obj["test_files"]) == 0:
        print("\tLanguage {} (fam: {}) has zero test files, skipping. ".format(l["language_name"], g))
        pprint(obj)
        continue

    filename = g.lower()+".yaml"
    with open(os.path.join(folder, filename), 'w') as f:
        yaml.dump(obj, f, sort_keys=True)



from cube.api import _load
from cube.version import __version__

import logging
logger = logging.getLogger('cube')

if logger.level == 0:
    logger.setLevel(logging.INFO)

log_handler = logging.StreamHandler()
log_formatter = logging.Formatter(fmt="%(asctime)s %(levelname)s: %(message)s", datefmt='%Y-%m-%d %H:%M:%S')
log_handler.setFormatter(log_formatter)


import sys, argparse


def fopen(filename, mode="r"):
    if sys.version_info[0] == 2:
        return open(filename, mode)
    else:
        if "b" in mode.lower():
            return open(filename, mode)
        else:
            return open(filename, mode, encoding="utf-8")


class ArgParser():
    def __init__(self):
        self.parser = argparse.ArgumentParser(description='Tagger ')
        self.parser.add_argument('--train', action='store', dest='train_file',
                                 help='Start building a tagger model')
        self.parser.add_argument('--patience', action='store', type=int, default=20, dest='patience',
                                 help='Number of epochs before early stopping (default=20)')
        self.parser.add_argument('--store', action='store', dest='store', help='Output base', default='data/model')
        self.parser.add_argument('--gpus', action='store', dest='gpus', type=int,
                                 help='How many GPUs to use (default=1)', default=1)
        self.parser.add_argument('--num-workers', action='store', dest='num_workers', type=int,
                                 help='How many dataloaders to use (default=4)', default=4)
        self.parser.add_argument('--batch-size', action='store', type=int, default=16, dest='batch_size',
                                 help='Batch size (default=16)')
        self.parser.add_argument('--debug', action='store_true', dest='debug',
                                 help='Do some standard stuff to debug the model')
        self.parser.add_argument('--resume', action='store_true', dest='resume', help='Resume training')
        self.parser.add_argument('--lm-model', action='store', dest='lm_model', default='transformer:xlm-roberta-base',
                                 help='What LM model to use (default=transformer:xlm-roberta-base)')
        self.parser.add_argument('--lm-device', action='store', dest='lm_device', default='cuda:0',
                                 help='Where to load LM (default=cuda:0)')
        self.parser.add_argument('--config', action='store', dest='config_file', help='Load config file')

    def __call__(self, *args, **kwargs):
        return self.parser.parse_args()



import sys
import torch
import numpy as np

sys.path.append('')
from typing import *
from abc import abstractmethod
from transformers import AutoModel, AutoTokenizer
from cube.io_utils.encodings import Encodings
from cube.io_utils.objects import Sentence
from cube.networks.lm import LMHelperLanguasito, LMHelperFT
from torch.utils.data.dataset import Dataset


class TokenCollate:
    def __init__(self):
        pass

    @abstractmethod
    def collate_fn(self, batch):
        pass

    @abstractmethod
    def get_embeddings_size(self) -> int:
        pass

    @abstractmethod
    def collate_fn_live(self, text, lang_id: int, batch_size: int):
        pass


class LanguasitoTokenizer:
    def __init__(self, no_space_language=False):
        self._no_space_language = no_space_language

    def __call__(self, text):
        if self._no_space_language:
            return [ch for ch in text]
        else:
            toks = []
            tok = ''
            for ch in text:
                if not ch.isalnum() or ch == ' ':
                    tok = tok.strip()
                    if len(tok) != 0:
                        toks.append(tok)
                        tok = ''
                    if ch != ' ':
                        toks.append(ch)
                else:
                    tok += ch
            if tok.strip() != '':
                toks.append(tok)

            return toks


def _make_example_from_raw(toks, iBatch, seq_len, overlap):
    batch = []
    num_batches = len(toks[0]) // seq_len
    if len(toks[0]) % seq_len != 0:
        num_batches += 1
    start = iBatch * seq_len
    stop = min(iBatch * seq_len + seq_len, len(toks[0]))
    current = toks[0][start:stop]
    left = max(0, start - overlap)
    right = min(len(toks[0]), stop + overlap)
    prev = toks[0][left:start]
    next = toks[0][stop:right]
    if len(toks) == 2:  # TF or languasito
        current_spa = toks[1][start:stop]
        prev_spa = toks[1][left:start]
        next_spa = toks[1][stop:right]
        prev = (prev, prev_spa)
        main = (current, current_spa)
        next = (next, next_spa)
    else:
        current_spa = toks[2][start:stop]
        prev_spa = toks[2][left:start]
        next_spa = toks[2][stop:right]
        current_ids = toks[1][start:stop]
        prev_ids = toks[1][left:start]
        next_ids = toks[1][stop:right]
        prev = (prev, prev_ids, prev_spa)
        main = (current, current_ids, current_spa)
        next = (next, next_ids, next_spa)

    # if len(prev)==0:
    #    prev=['']
    # if len(next)==0:
    #    next=['']
    example = {'prev': prev, 'main': main, 'next': next}
    return example


class TokenDatasetLive(Dataset):
    def __init__(self, raw_text, pretokenize_func, seq_len=300, overlap=100):
        self._tokenize = pretokenize_func
        if raw_text[-1] != ' ':
            raw_text += ' '
        self._toks = pretokenize_func(raw_text)  # self._tokenize(raw_text)
        self._seq_len = seq_len
        self._overlap = overlap
        self._num_examples = len(self._toks[0]) // seq_len
        if len(self._toks[0]) % seq_len != 0:
            self._num_examples += 1

    def __len__(self):
        return self._num_examples

    def __getitem__(self, item):
        return _make_example_from_raw(self._toks, item, self._seq_len, self._overlap)


class TokenCollateFTLanguasito(TokenCollate):
    def __init__(self, encodings: Encodings, lm_model: str = None, lm_device: str = 'cuda:0', no_space_lang=False,
                 lang_id=None):
        self._encodings = encodings
        # from languasito.utils import LanguasitoTokenizer
        self._tokenizer = LanguasitoTokenizer(no_space_language=no_space_lang)
        self._emb_size = 0
        self._lm_model = lm_model
        self._lm_device = lm_device
        self._lang_id = lang_id
        self.max_seq_len = -1
        parts = lm_model.split(':')
        if parts[0] == 'fasttext':
            self._lm_helper = LMHelperFT(device=lm_device, model=parts[1])
            self._emb_size = [300]
        elif parts[0] == 'languasito':
            self._lm_helper = LMHelperLanguasito(device=lm_device, model=parts[1])
            self._emb_size = [512]
        else:
            print("UserWarning: unsupported LM type for tokenizer")

    def collate_fn(self, batch):
        START = 0
        END = 2
        PAD = 1
        max_x = 0
        x_input = []
        x_input_spa = []
        x_lang = []
        y_output = []
        y_offset = []
        y_len = []
        x_text = []
        x_lang_word = []
        x_sent_len = []

        x_word_embeddings = []
        a_word_len = []
        for example in batch:
            for qq in ['prev', 'main', 'next']:
                sent = example[qq]
                # toks, ids = self._tokenize(sent.text)
                if self._lang_id is None:
                    toks = self._tokenizer(sent.text)
                    l_id = sent.lang_id
                else:
                    toks, spa = sent
                    l_id = self._lang_id
                for word in toks:
                    a_word_len.append(len(word))
                    x_lang_word.append(l_id)

        x_word_len = np.array(a_word_len, dtype=np.long)
        max_word_len = np.max(x_word_len)
        x_word_masks = np.zeros((x_word_len.shape[0], max_word_len), dtype=np.float)
        x_word = np.zeros((x_word_len.shape[0], max_word_len), dtype=np.long)
        x_word_case = np.zeros((x_word_len.shape[0], max_word_len), dtype=np.long)
        c_word = 0
        for example in batch:
            sz = 0
            for qq in ['prev', 'main', 'next']:
                sent = example[qq]
                # toks, ids = self._tokenize(sent.text)
                if self._lang_id is None:
                    toks = self._tokenizer(sent.text)
                else:
                    toks, spa = sent
                lst = toks
                sz += len(lst)
                for word in lst:
                    for iChar in range(len(word)):
                        x_word_masks[c_word, iChar] = 1
                        ch = word[iChar]
                        if ch.lower() == ch.upper():  # symbol
                            x_word_case[c_word, iChar] = 1
                        elif ch.lower() != ch:  # upper
                            x_word_case[c_word, iChar] = 2
                        else:  # lower
                            x_word_case[c_word, iChar] = 3
                        ch = ch.lower()
                        if ch in self._encodings.char2int:
                            x_word[c_word, iChar] = self._encodings.char2int[ch]
                        else:
                            x_word[c_word, iChar] = self._encodings.char2int['<UNK>']
                    c_word += 1
            x_sent_len.append(sz)

        for example in batch:
            current_sentence = example['main']
            prev_sentence = example['prev']
            next_sentence = example['next']
            if self._lang_id is None:
                x_lang.append(current_sentence.lang_id + 1)
            else:
                x_lang.append(self._lang_id + 1)
            # toks, ids = self._tokenize(prev_sentence.text)
            if self._lang_id is None:
                toks, spa = self.get_tokens(prev_sentence.text)
            else:
                toks, spa = prev_sentence
            x_prev = toks
            x_prev_spa = spa
            # toks, ids = self._tokenize(next_sentence.text)
            if self._lang_id is None:
                toks, spa = self.get_tokens(next_sentence.text)
            else:
                toks, spa = next_sentence
            x_next = toks
            x_next_spa = spa
            y_offset.append(len(x_prev))
            # c_toks, ids = self._tokenize(current_sentence.text)
            if self._lang_id is None:
                c_toks, c_spa = self.get_tokens(current_sentence.text)
            else:
                c_toks, c_spa = current_sentence
            x_main = c_toks
            x_main_spa = c_spa
            y_len.append(len(x_main))
            x_len = len(x_prev) + len(x_main) + len(x_next)
            x_input.append([x_prev, x_main, x_next])
            x_input_spa.append([x_prev_spa, x_main_spa, x_next_spa])
            x_text.append(c_toks)
            if self._lang_id is None:
                y_output.append(self._get_targets(current_sentence))
            else:
                y_output.append(np.zeros(len(c_toks)))

            if x_len > max_x:
                max_x = x_len

        x_for_emb = []
        for example in x_input:
            toks = example[0] + example[1] + example[2]
            x_for_emb.append(toks)

        x_emb = self._lm_helper.apply_raw(x_for_emb)
        max_len = max([len(x) for x in x_emb])
        x_out = np.zeros((len(x_emb), max_len, self._emb_size[0]), dtype=np.float)
        x_out_spa = np.zeros((len(x_emb), max_len), dtype=np.long)
        for ii in range(x_out.shape[0]):
            for jj in range(x_out.shape[1]):
                if jj < len(x_emb[ii]):
                    x_out[ii, jj, :] = x_emb[ii][jj]

            pos = 0
            spa = x_input_spa[ii][0]
            for jj in range(len(spa)):
                x_out_spa[ii, pos] = spa[jj]
                pos += 1
            spa = x_input_spa[ii][1]
            for jj in range(len(spa)):
                x_out_spa[ii, pos] = spa[jj]
                pos += 1
            spa = x_input_spa[ii][2]
            for jj in range(len(spa)):
                x_out_spa[ii, pos] = spa[jj]
                pos += 1

        y_out = np.zeros((x_out.shape[0], x_out.shape[1]), dtype=np.long)
        for ii in range(x_out.shape[0]):
            for jj in range(y_len[ii]):
                index = y_offset[ii] + jj
                y_out[ii, index] = y_output[ii][jj]
        x_out = torch.tensor(x_out)
        x_lang = torch.tensor(x_lang)
        y_out = torch.tensor(y_out)
        y_offset = torch.tensor(y_offset)
        y_len = torch.tensor(y_len)
        x_word = torch.tensor(x_word)
        x_word_case = torch.tensor(x_word_case)
        x_word_masks = torch.tensor(x_word_masks)
        x_word_len = torch.tensor(x_word_len)
        x_lang_word = torch.tensor(x_lang_word)
        x_sent_len = torch.tensor(x_sent_len)
        x_input_spa = torch.tensor(x_out_spa)

        return {'x_input': [x_out], 'x_input_spa': x_input_spa, 'x_word_char': x_word, 'x_word_case': x_word_case,
                'x_word_masks': x_word_masks, 'x_word_len': x_word_len, 'x_word_lang': x_lang_word, 'x_text': x_text,
                'x_lang': x_lang, 'y_output': y_out, 'y_offset': y_offset, 'y_len': y_len, 'x_sent_len': x_sent_len}

    def _get_targets(self, sentence: Sentence):
        text = sentence.text
        toks = self._tokenizer(text)

        targets = [0 for _ in range(len(toks))]
        iToken = 0
        cl = 0
        for ii in range(len(targets)):
            target = 1  # nothing
            cl += len(toks[ii].replace(' ', ''))
            if cl == len(sentence.tokens[iToken].text.replace(' ', '')):
                iToken += 1
                cl = 0
                target = 2  # token
                if len(sentence.tokens[iToken - 1].words) > 1:
                    target = 3  # multiword token
            if iToken == len(sentence.tokens):
                target = 4  # sentence end (+token)
                for tt in range(ii, len(targets)):
                    targets[ii] = target
                break
            targets[ii] = target
        return targets

    def get_tokens(self, text):
        toks = self._tokenizer(text)
        spa = [0 for _ in range(len(toks))]
        t_pos = 0
        for ii in range(len(toks)):
            t_pos += len(toks[ii])
            if t_pos < len(text) and text[t_pos] == ' ':
                spa[ii] = 2
                while t_pos < len(text) and text[t_pos] == ' ':
                    t_pos += 1
            else:
                spa[ii] = 1
        return toks, spa

    def get_embeddings_size(self) -> List[int]:
        return self._emb_size

    def __getstate__(self):
        state = self.__dict__.copy()
        # Don't pickle baz
        if "_lm_helper" in state:
            del state["_lm_helper"]
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        parts = self._lm_model.split(':')
        if parts[0] == 'fasttext':
            self._lm_helper = LMHelperFT(device=self._lm_device, model=parts[1])
            self._emb_size = [300]
        elif parts[0] == 'languasito':
            self._lm_helper = LMHelperLanguasito(device=self._lm_device, model=parts[1])
            self._emb_size = [512]


class TokenCollateHF(TokenCollate):
    def __init__(self, encodings: Encodings, lm_device, lm_model=None, no_space_lang=False, lang_id=None):
        if lm_model is None:
            lm_model = 'xlm-roberta-base'
        self._encodings = encodings  # this is currently not used - we keep it for future development
        self._pretokenizer = LanguasitoTokenizer(no_space_language=no_space_lang)
        self._tokenizer = AutoTokenizer.from_pretrained(lm_model)
        self._lm = AutoModel.from_pretrained(lm_model, output_hidden_states=True)
        self._lm.eval()
        self._lm_device = lm_device
        self._lm.to(lm_device)
        self._no_space = no_space_lang
        tmp = self._lm(torch.tensor([[100]], device=lm_device))
        h_state_size = tmp['hidden_states'][0].shape[-1]
        self._emb_size = [h_state_size for _ in range(len(tmp['hidden_states']))]
        self._lang_id = lang_id

        self.max_seq_len = self._tokenizer.model_max_length

    def get_tokens(self, text):
        toks, ids = self._tokenize(text)
        spa = [0 for _ in range(len(toks))]
        t_pos = 0
        for ii in range(len(toks)):
            t_pos += len(toks[ii].replace('▁', ''))
            if t_pos < len(text) and text[t_pos] == ' ':
                spa[ii] = 2
                while t_pos < len(text) and text[t_pos] == ' ':
                    t_pos += 1
            else:
                spa[ii] = 1
        return toks, ids, spa

    def collate_fn(self, batch):
        START = 0
        END = 2
        PAD = 1
        max_x = 0
        x_input = []
        x_input_spa = []
        x_lang = []
        y_output = []
        y_offset = []
        y_len = []
        x_text = []
        x_spa = []
        x_lang_word = []
        x_sent_len = []

        x_word_embeddings = []
        a_word_len = []
        for example in batch:
            for qq in ['prev', 'main', 'next']:
                sent = example[qq]
                if self._lang_id is None:
                    l_id = sent.lang_id
                    text = sent.text
                    toks, ids, spa = self.get_tokens(text)
                else:
                    l_id = self._lang_id
                    if len(sent) == 3:
                        toks, ids, spa = sent
                    else:
                        toks, ids, spa = [], [], []
                for word in toks:
                    a_word_len.append(len(word))
                    x_lang_word.append(l_id)
        x_word_len = np.array(a_word_len, dtype=np.long)
        max_word_len = np.max(x_word_len)
        x_word_masks = np.zeros((x_word_len.shape[0], max_word_len), dtype=np.float)
        x_word = np.zeros((x_word_len.shape[0], max_word_len), dtype=np.long)
        x_word_case = np.zeros((x_word_len.shape[0], max_word_len), dtype=np.long)
        c_word = 0
        for example in batch:
            sz = 0
            for qq in ['prev', 'main', 'next']:
                sent = example[qq]
                if self._lang_id is None:
                    l_id = sent.lang_id
                    toks, ids, spa = self.get_tokens(sent.text)
                else:
                    l_id = self._lang_id
                    if len(sent) == 3:
                        toks, ids, spa = sent
                    else:
                        toks, ids, spa = [], [], []

                lst = toks
                sz += len(lst)
                for word in lst:
                    for iChar in range(len(word)):
                        x_word_masks[c_word, iChar] = 1
                        ch = word[iChar]
                        if ch.lower() == ch.upper():  # symbol
                            x_word_case[c_word, iChar] = 1
                        elif ch.lower() != ch:  # upper
                            x_word_case[c_word, iChar] = 2
                        else:  # lower
                            x_word_case[c_word, iChar] = 3
                        ch = ch.lower()
                        if ch in self._encodings.char2int:
                            x_word[c_word, iChar] = self._encodings.char2int[ch]
                        else:
                            x_word[c_word, iChar] = self._encodings.char2int['<UNK>']
                    c_word += 1
            x_sent_len.append(sz)

        for example in batch:
            current_sentence = example['main']
            prev_sentence = example['prev']
            next_sentence = example['next']
            if self._lang_id is None:
                x_lang.append(current_sentence.lang_id + 1)
            else:
                x_lang.append(self._lang_id + 1)
            if self._lang_id is None:
                toks, ids, spa = self.get_tokens(prev_sentence.text)
            else:
                if len(prev_sentence) == 3:
                    toks, ids, spa = prev_sentence
                else:
                    toks, ids, spa = [], [], []
            x_prev = ids
            x_prev_spa = spa
            if self._lang_id is None:
                toks, ids, spa = self.get_tokens(next_sentence.text)
            else:
                if len(next_sentence) == 3:
                    toks, ids, spa = next_sentence
                else:
                    toks, ids, spa = [], [], []
            x_next = ids
            x_next_spa = spa
            y_offset.append(len(x_prev))
            if self._lang_id is None:
                c_toks, ids, c_spa = self.get_tokens(current_sentence.text)
            else:
                if len(current_sentence) == 3:
                    c_toks, ids, c_spa = current_sentence
                else:
                    c_toks, ids, c_spa = [], [], []

            x_main = ids
            x_main_spa = c_spa
            y_len.append(len(x_main))
            x_len = len(x_prev) + len(x_main) + len(x_next)
            x_input.append([x_prev, x_main, x_next])
            x_input_spa.append([x_prev_spa, x_main_spa, x_next_spa])
            x_text.append(c_toks)
            x_spa.append(c_spa)
            if self._lang_id is None:
                y_output.append(self._get_targets(current_sentence))
            else:
                y_output.append([0 for _ in range(len(c_toks))])
            if x_len > max_x:
                max_x = x_len

        x_out = np.ones((len(batch), max_x), dtype=np.long) * PAD
        x_out_spa = np.zeros((len(batch), max_x), dtype=np.long)
        for ii in range(len(batch)):
            # x_out[ii, 0] = START
            pos = 0
            x = x_input[ii][0]
            x_spa = x_input_spa[ii][0]
            for jj in range(len(x)):
                x_out[ii, pos] = x[jj]
                x_out_spa[ii, pos] = x_spa[jj]
                pos += 1
            x = x_input[ii][1]
            x_spa = x_input_spa[ii][1]
            for jj in range(len(x)):
                x_out[ii, pos] = x[jj]
                x_out_spa[ii, pos] = x_spa[jj]
                pos += 1
            x = x_input[ii][2]
            x_spa = x_input_spa[ii][2]
            for jj in range(len(x)):
                x_out[ii, pos] = x[jj]
                x_out_spa[ii, pos] = x_spa[jj]
                pos += 1
            # x_out[ii, pos] = END

        y_out = np.zeros((x_out.shape[0], x_out.shape[1]), dtype=np.long)
        for ii in range(x_out.shape[0]):
            for jj in range(y_len[ii]):
                index = y_offset[ii] + jj
                y_out[ii, index] = y_output[ii][jj]
        x_out = torch.tensor(x_out, device=self._lm_device)
        x_lang = torch.tensor(x_lang)
        y_out = torch.tensor(y_out)
        x_input_spa = torch.tensor(x_out_spa, dtype=torch.long)
        y_offset = torch.tensor(y_offset)
        y_len = torch.tensor(y_len)
        x_word = torch.tensor(x_word)
        x_word_case = torch.tensor(x_word_case)
        x_word_masks = torch.tensor(x_word_masks)
        x_word_len = torch.tensor(x_word_len)
        x_lang_word = torch.tensor(x_lang_word)
        x_sent_len = torch.tensor(x_sent_len)
        with torch.no_grad():
            if x_out.size()[1] > self.max_seq_len:
                # print()
                # print(x_out.size())
                # hack to skip batch if len is to big
                # we cannot return none because pytorch lightning will complain, so we set x_input = None and check
                # it in the train_step. TODO check long string in processing input
                return {'x_input': None, 'x_input_spa': x_input_spa, 'x_word_char': x_word, 'x_word_case': x_word_case,
                        'x_word_masks': x_word_masks,
                        'x_word_len': x_word_len, 'x_word_lang': x_lang_word, 'x_text': x_text, 'x_lang': x_lang,
                        'y_output': y_out, 'y_offset': y_offset, 'y_len': y_len, 'x_sent_len': x_sent_len}
            x_out = self._lm(x_out)['hidden_states']
            x_out = [t.detach() for t in x_out]
        return {'x_input': x_out, 'x_input_spa': x_input_spa, 'x_word_char': x_word, 'x_word_case': x_word_case,
                'x_word_masks': x_word_masks,
                'x_word_len': x_word_len, 'x_word_lang': x_lang_word, 'x_text': x_text, 'x_lang': x_lang,
                'y_output': y_out, 'y_offset': y_offset, 'y_len': y_len, 'x_sent_len': x_sent_len}

    def _tokenize(self, text):
        if self._no_space:
            new_text = [ch for ch in text]
        else:
            new_text = self._pretokenizer(text)
        # print("\n" + ("_" * 50))
        # print(new_text)
        # print("_" * 50)
        toks = self._tokenizer.tokenize(new_text, is_split_into_words=True)
        ids = self._tokenizer(new_text, is_split_into_words=True)['input_ids'][1:-1]
        r_toks = []
        r_ids = []
        strip_text = text.replace(' ', '').replace('\n', '').replace('\r', '').replace('\t', '')
        strip_index = 0
        # if len(toks) != 0:  # empty text
        #     r_toks.append(toks[0])
        #     r_ids.append(ids[0])
        for ii in range(0, len(toks)):
            if toks[ii] != '▁':
                r_toks.append(toks[ii])
                r_ids.append(ids[ii])
        # if len(r_toks) > 509 or len(r_ids) > 509 or len(r_toks) < 1 or len(r_ids) < 1:
        #    print(f"\n>> text:[{text}] [{len(r_toks)}] [{len(r_ids)}]")
        f_toks = []
        for tok in r_toks:
            tok = tok.replace('▁', '')
            s_len = len(tok)
            f_toks.append(strip_text[strip_index:strip_index + s_len])
            strip_index += s_len

        return f_toks, r_ids

    def _get_targets(self, sentence: Sentence):
        text = sentence.text
        toks, ids = self._tokenize(text)
        toks = [tok.replace('▁', '') for tok in toks]
        targets = [0 for _ in range(len(toks))]
        iToken = 0
        cl = 0
        for ii in range(len(targets)):
            target = 1  # nothing
            cl += len(toks[ii])
            if cl == len(sentence.tokens[iToken].text):
                iToken += 1
                cl = 0
                target = 2  # token
                if len(sentence.tokens[iToken - 1].words) > 1:
                    target = 3  # multiword token
            if iToken == len(sentence.tokens):
                target = 4  # sentence end (+token)
                for tt in range(ii, len(targets)):
                    targets[ii] = target
                break
            targets[ii] = target
        return targets

    def get_embeddings_size(self) -> List[int]:
        return self._emb_size



# This script reads the results.json file and converts it to a results.md file

import json, collections

def extract_language_from_test_file(file):
    # ex: "/home/ubuntu/ud-treebanks-v2.2/UD_Afrikaans-AfriBooms/af_afribooms-ud-test.txt"
    parts = file.split("/")
    dir = parts[-2].replace("UD_","")
    dir = dir[:dir.find("-")]
    return dir.replace("_"," ")

all = json.load(open("results.json","r"))

lall = {}
for model, elem in all.items():
    language = extract_language_from_test_file(elem["test_file"])
    print(language)
    if language not in lall:
        lall[language] = {}    
    lall[language][model] = elem
all = collections.OrderedDict(sorted(lall.items()))

header =  "|Language|Model|Token|Sentence|UPOS|XPOS|AllTags|Lemmas|UAS|LAS|\n"
header += "|--------|-----|:---:|:------:|:--:|:--:|:-----:|:----:|:-:|:-:|\n"
rows = []
for language, langdict in all.items():
    lrow = "|"+language+"|\n"
    rows.append(lrow)
    langdict = collections.OrderedDict(sorted(langdict.items()))
    for model, elem in langdict.items():
        row = "| |"
        row += model + "|"        
        row += str(round(elem["metrics"]["Tokens"]["f1"],2)) + "|"
        row += str(round(elem["metrics"]["Sentences"]["f1"],2)) + "|"
        row += str(round(elem["metrics"]["UPOS"]["f1"],2)) + "|"
        row += str(round(elem["metrics"]["XPOS"]["f1"],2)) + "|"
        row += str(round(elem["metrics"]["AllTags"]["f1"],2)) + "|"
        row += str(round(elem["metrics"]["Lemmas"]["f1"],2)) + "|"
        row += str(round(elem["metrics"]["UAS"]["f1"],2)) + "|"
        row += str(round(elem["metrics"]["LAS"]["f1"],2)) + "|\n"
        rows.append(row)
        
with open("results.md","w",encoding="utf8") as f:
    f.write(header)
    for row in rows:
        f.write(row)